# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.enum_option_id import ArrayOfEnumOptionId
from pylero.enum_option_id import EnumOptionId
from pylero.test_step import ArrayOfTestStep
from pylero.test_step import TestStep


class TestSteps(BasePolarion):
    """Object to handle the Polarion WSDL tns3:TestSteps class

    Attributes:
        keys (ArrayOfEnumOptionId)
        steps (ArrayOfTestStep)
"""
    _cls_suds_map = {"keys":
                     {"field_name": "keys",
                      "is_array": True,
                      "cls": EnumOptionId,
                      "arr_cls": ArrayOfEnumOptionId,
                      "inner_field_name": "EnumOptionId",
                      "enum_id": "testing/test-step-keys"},
                     "steps":
                     {"field_name": "steps",
                      "is_array": True,
                      "cls": TestStep,
                      "arr_cls": ArrayOfTestStep,
                      "inner_field_name": "TestStep"},
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "test_management_client"
    _obj_struct = "tns3:TestSteps"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.enum_option_id import EnumOptionId
from pylero.work_item import _WorkItem


class BuildLinkedWorkItem(BasePolarion):
    """Object to handle the Polarion WSDL tns2:BuildLinkedWorkItem class

    Attributes:
        revision (string)
        role (EnumOptionId)
        work_item (WorkItem)
"""
    _cls_suds_map = {"revision": "revision",
                     "role":
                     {"field_name": "role",
                      "cls": EnumOptionId,
                      "enum_id": "workitem-link-role"},
                     "work_item":
                     {"field_name": "workItem",
                      "cls": _WorkItem},
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns2:BuildLinkedWorkItem"


class ArrayOfBuildLinkedWorkItem(BasePolarion):
    _obj_client = "builder_client"
    _obj_struct = "tns2:ArrayOfBuildLinkedWorkItem"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class PriorityOptionId(BasePolarion):
    """Object to handle the Polarion WSDL tns5:PriorityOptionId class

    Attributes:
        id (string)
"""
    _cls_suds_map = {"id": "id",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _id_field = "id"
    _obj_client = "builder_client"
    _obj_struct = "tns5:PriorityOptionId"

    def __init__(self, id=None, uri=None,
                 suds_object=None):
        """PriorityOptionID Constructor

        Args:
            id: value of the priority
            uri: the SubterraURI of the priority
            suds_object: the Polarion Plan object

        Returns:
            None

        References:
            _WorkItem.priority
        """
        super(self.__class__, self).__init__(id, suds_object)


class ArrayOfPriorityOptionId(BasePolarion):
    _obj_client = "builder_client"
    _obj_struct = "tns5:ArrayOfPriorityOptionId"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os

import suds
from pylero._compatible import basestring
from pylero._compatible import classmethod
from pylero._compatible import object  # noqa
from pylero._compatible import range
from pylero.base_polarion import BasePolarion
from pylero.base_polarion import tx_wrapper
from pylero.build import Build  # noqa: F401
from pylero.custom import ArrayOfCustom
from pylero.custom import Custom
from pylero.custom_field_type import CustomFieldType
from pylero.document import Document
from pylero.enum_custom_field_type import EnumCustomFieldType
from pylero.enum_option_id import ArrayOfEnumOptionId
from pylero.enum_option_id import EnumOptionId
from pylero.exceptions import PyleroLibException
from pylero.plan import Plan  # NOQA
from pylero.project import Project
from pylero.test_record import ArrayOfTestRecord
from pylero.test_record import TestRecord
from pylero.test_run_attachment import ArrayOfTestRunAttachment
from pylero.test_run_attachment import TestRunAttachment
from pylero.text import Text
from pylero.user import User
from pylero.work_item import _WorkItem
from pylero.work_item import TestCase
# Build is used in custom fields.
# Plan is used in custom fields.


def generate_description(test_run, test_case, test_record):
    tr_html = "<b>Test Run:</b> <span id=\"link\" class=    " \
              "\"polarion-rte-link\" data-type=\"testRun\" " \
              "data-item-id=\"{0}\" data-option-id=\"long\">" \
              "</span><br/>".format(test_run.test_run_id)
    tc_html = "<b>Test Case:</b> <span id=\"link\" class=" \
              "\"polarion-rte-link\" data-type=\"workItem\" " \
              "data-item-id=\"{0}\" data-option-id=\"long\"></span>" \
              "<br/>".format(test_case.work_item_id)
    table_cell_style = "style=\"text-align: left; padding: 10px; " \
                       "vertical-align: top; background-color: #ffffff;\""
    table_row_style = "style=\"border-bottom: 1px solid #f0f0f0;\""
    columns = ["", "#", "<span title=\"Step\">Step</span>",
               "<span title=\"Expected Result\">Expected Result</span>",
               "Actual Result"]
    test_step_results = {"passed": "<span title=\"Results met expected results"
                                   "\"><span style=\"white-space:nowrap;\"><im"
                                   "g src=\"/polarion/icons/default/enums/test"
                                   "run_status_passed.png\" style=\"vertical-a"
                                   "lign:text-bottom;border:0px;margin-right:2"
                                   "px;\" class=\"polarion-no-style-cleanup\"/"
                                   "></span></span>",
                         "failed": "<span title=\"Results did not meet expecte"
                                   "d results\"><span style=\"white-space:nowr"
                                   "ap;\"><img src=\"/polarion/icons/default/e"
                                   "nums/testrun_status_failed.png\" style=\"v"
                                   "ertical-align:text-bottom;border:0px;margi"
                                   "n-right:2px;\" class=\"polarion-no-style-c"
                                   "leanup\"/></span></span>",
                         "blocked": "<span title=\"Errors in the product preve"
                                    "nted test from being executed\"><span sty"
                                    "le=\"white-space:nowrap;\"><img src=\"/po"
                                    "larion/icons/default/enums/testrun_status"
                                    "_blocked.png\" style=\"vertical-align:tex"
                                    "t-bottom;border:0px;margin-right:2px;\" c"
                                    "lass=\"polarion-no-style-cleanup\"/>"
                                    "</span></span>"}
    table_header = "<table class=\"polarion-no-style-cleanup\" style=\"border"\
                   "-collapse: collapse;\"><tr style=\"text-align: left; " \
                   "white-space: nowrap; color: #757575; border-bottom: 1px " \
                   "solid #d2d7da; background-color: #ffffff;\">{0}</tr>" \
        .format("".join(["<th {0}>{1}</th>".format(table_cell_style, column)
                         for column in columns]))
    verdict = "</table><table style=\"margin-bottom: 15px; ;border-collapse: "\
              "collapse; width:100%; ;margin-top: 13px;\" class=\"polarion-no"\
              "-style-cleanup\"><tr><th style=\"width: 80%; text-align: left;"\
              " background-color: #ffffff;\">Test Case Verdict:</th></tr><tr>"\
              "<td style=\"vertical-align: top;\"><span style=\"font-weight: "\
              "bold;\"><span style=\"color: #C30000;\"><span title=\"Results "\
              "did not meet expected results\"><span style=\"white-space:" \
              "nowrap;\"><img src=\"/polarion/icons/default/enums/testrun_" \
              "status_failed.png\" style=\"vertical-align:text-bottom;border:"\
              "0px;margin-right:2px;\" class=\"polarion-no-style-cleanup\"/>" \
              "</span>Failed</span></span></span><span> {0}</span></td></tr>" \
              "</table>" \
        .format(test_record.comment)
    table_rows = ""
    for step in range(len(test_record.test_step_results)):
        table_rows += "<tr {0}>" \
                      "<td {1}>{2}</td>" \
                      "<td {1}>{3}</td>" \
                      "<td {1}>{4}</td>" \
                      "<td {1}>{5}</td>" \
                      "<td {1}>{6}</td>" \
                      "</tr>".format(table_row_style,
                                     table_cell_style,
                                     test_step_results.
                                     get(test_record.test_step_results[step]
                                         .result),
                                     step + 1,
                                     test_case.test_steps.steps[step].values[0]
                                     .content,
                                     test_case.test_steps.steps[step].values[1]
                                     .content,
                                     test_record.test_step_results[step]
                                     .comment)
    content = tr_html + tc_html + table_header + table_rows + verdict

    return content


def create_incident_report(test_run, test_record, test_case):
    project_id = test_run.project_id
    status = 'open'
    project = Project(project_id)
    tconf = project.get_tests_configuration()
    defectWorkItemType = tconf.defect_work_item_type
    title = 'Failed: ' + test_case.title
    description = generate_description(test_run, test_case, test_record)
    kwarg_dict = {}

    for prop in tconf.fields_to_copy_from_test_case_to_defect.property:
        kwarg_dict[prop.value] = getattr(test_case, prop.key)
    for prop in tconf.fields_to_copy_from_test_run_to_linked_defect.property:
        kwarg_dict[prop.value] = getattr(test_run, prop.key)

    incident_report = _WorkItem.create(project_id, defectWorkItemType, title,
                                       description, status, **kwarg_dict)
    incident_report.add_linked_item(test_case.work_item_id, "triggered_by")
    return incident_report.work_item_id


class TestRun(BasePolarion):
    """Object to manage the Polarion Test Management WS tns3:TestRun

    Attributes:
        attachments (list of TestRunAttachments)
        author (User): user object of the Test Run Author
        created (datetime)
        document (Module)
        finished_on (datetime)
        group_id
        is_template (bool): indicates if the TestRun object is a Template
        keep_in_history (bool)
        location
        project (Project)
        query (str): The Polarion query that the TestRun objects are based on.
        records (list of TestRecord objects)
        select_test_cases_by (EnumOptionId):
            The test cases can be:
                AUTOMATED_PROCESS
                DYNAMIC_QUERY
                DYNAMIC_LIVEDOC
                MANUAL_SELECTION
                STATIC_QUERY
                STATIC_LIVEDOC

        status
        summary_defect (_WorkItem)
        template (TestRun): template that the TestRun is based on.
        test_run_id (str): Unique identifier of the test run within the
                           project
        type
        updated
        custom_fields
    """
    _cls_suds_map = {
        "attachments":
            {"field_name": "attachments",
             "is_array": True,
             "cls": TestRunAttachment,
             "arr_cls": ArrayOfTestRunAttachment,
             "inner_field_name": "TestRunAttachment"},
        "author":
            {"field_name": "authorURI",
             "cls": User,
             "named_arg": "uri",
             "sync_field": "uri"},
        "created": "created",
        "document":
            {"field_name": "document",
             "cls": Document},
        "finished_on": "finishedOn",
        "group_id": "groupId",
        "test_run_id": "id",
        "is_template": "isTemplate",
        "keep_in_history": "keepInHistory",
        "location": "location",
        "project_id":
            {"field_name": "projectURI",
             "cls": Project,
             "named_arg": "uri",
             "sync_field": "uri"},
        "query": "query",
        "_records":
            {"field_name": "records",
             "is_array": True,
             "cls": TestRecord,
             "arr_cls": ArrayOfTestRecord,
             "inner_field_name": "TestRecord"},
        "select_test_cases_by":
            {"field_name": "selectTestCasesBy",
             "cls": EnumOptionId,
             "enum_id": "testrun-selectTestCasesBy"},
        "status":
            {"field_name": "status",
             "cls": EnumOptionId,
             "enum_id": "testing/testrun-status"},
        "summary_defect":
            {"field_name": "summaryDefectURI",
             "cls": _WorkItem,
             "named_arg": "uri",
             "sync_field": "uri"},
        "title": "title",
        "template":
            {"field_name": "templateURI",
             "named_arg": "uri",
             "sync_field": "uri"},
        "type":
            {"field_name": "type",
             "cls": EnumOptionId,
             "enum_id": "testing/testrun-type"},
        "updated": "updated",
        # the custom field attribute has been changed to be a protected attr.
        # All interaction with custom fields should be done directly with the
        # derived attribute.
        "_custom_fields":
            {"field_name": "customFields",
             "is_array": True,
             "cls": Custom,
             "arr_cls": ArrayOfCustom,
             "inner_field_name": "Custom"},
        "uri": "_uri",
        "_unresolvable": "_unresolvable"}
    _id_field = "test_run_id"
    _obj_client = "test_management_client"
    _obj_struct = "tns3:TestRun"
    _custom_field_cache = {}

    @property
    def records(self):
        """ function to return all the test records of a TestRun.
        The records array for dynamic queries/documents only includes executed
        records. This returns the unexecuted ones as well.

        Args:
            None

        Returns:
            list of TestRecords
        """
        self._verify_obj()
        # if the type is not dynamic then all the cases are in the _records
        # attribute. If they are dynamic, they have to be gotten
        if "dynamic" not in self.select_test_cases_by:
            return self._records
        if "Doc" in self.select_test_cases_by:
            cases = self.document.get_work_items(None, True)
        elif "Query" in self.select_test_cases_by:
            cases = _WorkItem.query(
                self.query + " AND project.id:" + self.project_id)
        else:
            raise PyleroLibException("Only Test Runs based on Docs or"
                                     " Queries can be dynamic")
        executed_ids = [rec.test_case_id for rec in self._records]
        test_recs = self._records
        for case in cases:
            if case.work_item_id not in executed_ids \
                    and case.type != "heading":
                test_recs.append(
                    TestRecord(self.project_id, case.work_item_id))
        return test_recs

    @records.setter
    def records(self, val):
        self._records = val

    @classmethod
    @tx_wrapper
    def create(cls, project_id, test_run_id=None, template=None, title=None,
               **kwargs):
        """class method create for creating a new test run in Polarion

        Args:
            project_id (string): the Polarion project to create the test run
                                 in
            test_run_id (string): the unique identifier for the test run,
                                  If the Enable Generated Test Run IDs option
                                  is marked in the configuration, it will
                                  ignore this field. If it is None, the title
                                  will be used.
            template (string): the id of the template to base the test run on.
                               cannot be None. Because of the existing order of
                               params before adding title, it is given a
                               default value of None, which will cause an error
                               if it is not populated
            title (string): To create a test run with a title. In this case,
                            if the id is not populated it will be autogen.
                            If it is populated, it will use that.
            **kwargs: keyword arguments. Test run attributes can be passed in
                      to be set upon creation. Required fields must be passed
                      in

        Returns:
            The created TestRun object

        References:
            test_management.createTestRun
        """
        if not template:
            raise PyleroLibException("Template is required")
        if title:
            uri = cls.session.test_management_client.service.\
                createTestRunWithTitle(
                    project_id, test_run_id or title, title, template)
        else:
            uri = cls.session.test_management_client.service.createTestRun(
                project_id, test_run_id, template)
        if uri:
            run = cls(uri=uri)
            run.verify_params(**kwargs)
            for field in kwargs:
                setattr(run, field, kwargs[field])
            run.update()
            return run
        else:
            raise PyleroLibException("Test Run was not created")

    @classmethod
    @tx_wrapper
    def create_template(cls, project_id, template_id,
                        parent_template_id="Empty",
                        select_test_cases_by=None, query=None,
                        doc_with_space=None, **kwargs):  # , test_case_ids=[]):
        # see comment below regarding test_case)ids.
        """class method create_template for creating a new template in Polarion

        Args:
            project_id (string): the Polarion project to create the test run
                                 in
            template_id (string): the unique identifier for the template
            parent_template_id: the template that this is based on
                                Default: "Empty"
            select_test_cases_by: the method used to choose test cases
                                  NOTE: It is currently not possible to select
                                  test cases manually via the API.
                                  Default: None
            query: the Lucene query, for query methods, default None
            doc_with_space: the space/doc_name, for document methods
                            default: None
            **kwargs: keyword arguments. Test run attributes can be passed in
                      to be set upon creation. Required fields must be passed
                      in

        Returns:
            The created TestRun Template object

        References:
            test_management.createTestRun
        """
        tr = cls.create(project_id, template_id, parent_template_id, **kwargs)
        tr.is_template = True
        if select_test_cases_by:
            tr.select_test_cases_by = select_test_cases_by
        elif doc_with_space:
            tr.select_test_cases_by = "dynamicLiveDoc"
        elif query:
            tr.select_test_cases_by = "dynamicQueryResult"

        # you can have both a query and a document, through which the query
        # qualifies the doc.
        if query:
            tr.query = query
        if doc_with_space:
            tr.document = Document(project_id, doc_with_space)
# TODO: This should work as soon as Polarion implements Change Request REQ-6334
        #        if test_case_ids:
        #            for test_case_id in test_case_ids:
        #                tr.add_test_record_by_object(testrec.TestRecord(
        #                                            test_case_id,
        #                                            project_id))

        tr.update()
        return TestRun(tr.test_run_id, project_id=project_id)

    @classmethod
    def search(cls, query, fields=["test_run_id"], sort="test_run_id",
               limit=-1, search_templates=False, project_id=None):
        """class method search executes the given query and returns the results

        Args:
            query: the Polarion query used to find test runs
            fields:  test run fields that should be initialized,
                     all other fields will be null.
                     Field names are from the object's attributes and not
                     the Polarion field names, default ["test_run_id"]. If you
                     want all of the fields to be returned, pass [] for this
                     parameter.
            sort: the field used to sort results, default is test_run_id
            limit (int): the maximum number of records to be returned, -1
                         for no limit, default -1.
            search_templates (bool): if set, searches the templates
                                     instead of the test runs, default False
            project_id: if set, searches the project id, else default project
        Returns:
            list of TestRun objects

        References:
            test_management.searchTestRunTemplates
            test_management.searchTestRunTemplatesLimited
            test_management.searchTestRunTemplatesWithFields
            test_management.searchTestRunTemplatesWithFieldsLimited
            test_management.searchTestRuns
            test_management.searchTestRunLimited
            test_management.searchTestRunsWithFields
            test_management.searchTestRunsWithFieldsLimited
        """

# The Polarion functions with limited seem to be the same as without limited
#    when -1 is passed in as limit. Because of this, the wrapper will not
#    implement the functions without limited.
        project_id = project_id or cls.default_project
        query += " AND project.id:%s" % (project_id)

        # The following line one purpose is to instantiate a TestRun and by
        # doing so setting all the class attribute (including 'customFields').
        TestRun(project_id=project_id)
        function_name = "search"
        p_sort = cls._cls_suds_map[sort] if not isinstance(
            cls._cls_suds_map[sort], dict) else \
            cls._cls_suds_map[sort]["field_name"]
        parms = [query, p_sort]
        if search_templates:
            function_name += "TestRunTemplates"
        else:
            function_name += "TestRuns"
        p_fields = cls._convert_obj_fields_to_polarion(fields)
        if p_fields:
            function_name += "WithFieldsLimited"
            parms.append(p_fields)
        parms.append(limit)
        test_runs = []
        results = getattr(cls.session.test_management_client.service,
                          function_name)(*parms)
        for suds_obj in results:
            tr = TestRun(suds_object=suds_obj)
            test_runs.append(tr)
        return test_runs

    def __init__(self, test_run_id=None, suds_object=None, project_id=None,
                 uri=None):
        """TestRun constructor.

        Args:
            test_run_id: when given, the object is populated with the
                         TestRuns. Requires project_id parameter
            suds_object: Polarion TestRun object. When given, the object
                         is populated by object data.
            project_id: the Polarion project that the Test Run is located
                        in. Required if test_run_id is passed in
            uri: the uri that references the Polarion TestRun

        Notes:
            Either test_run_id and project or suds_object or uri can be passed
            in or none of them. If none of the identifying parameters are
            passed in an empty object is created

        References:
            test_management.getTestRunById
            test_management.getTestRunByUri
        """
        self._add_custom_fields(project_id)
        super(self.__class__, self).__init__(test_run_id, suds_object)
        if test_run_id:
            if not project_id:
                raise PyleroLibException("When test_run_id is passed in, "
                                         "project_id is required")
            self._suds_object = self.session.test_management_client.service. \
                getTestRunById(project_id, test_run_id)
        elif uri:
            self._suds_object = self.session.test_management_client.service. \
                getTestRunByUri(uri)
        if test_run_id or uri:
            if getattr(self._suds_object, "_unresolvable", True):
                raise PyleroLibException(
                    "The Test Run {0} was not found.".format(test_run_id))

    def _fix_circular_refs(self):
        # a class can't reference itself as a class attribute so it is
        # defined after instatiation
        self._cls_suds_map["template"]["cls"] = self.__class__

    def _custom_field_types(self, field_type):
        """There are 4 types of custom fields in test runs:
        * built-in types (string, boolean, ...)
        * Pylero Text object (text)
        * Enum of existing type (@ prefix, i.e. enum:@user = Pylero User)
        * Enum, based on lookup table(i.e. enum:arch)
        The basic enums can get their valid values using the BasePolarion
        get_valid_field_values function.
        The existing type Enums, must validate by instantiating the object.

        Args;
            field_type - the field type passed in custom field enum_id
        Returns:
            None for base types, Text class for text types, the object to
            validate the enum for Enum objects or the key to validate the enum
            for basic enums
        """
        if field_type == "text":
            return Text
        # some custom types have a [] segment. Still unsure of how to handle
        # those specific attributes, but for now, this will ignore them
        field_type = field_type.split("[")[0]
        if field_type.startswith("@"):
            # an enum based on an object
            return [globals()[x] for x in globals()
                    if x.lower() == field_type[1:].lower()][0]
        else:
            # a regular enum
            return field_type

    @classmethod
    def get_defined_custom_field_types(cls, project_id):
        """Gets all the custom fields defined for the specified project.
        the custom fields are all either of type CustomFieldType or
        EnumCustomFieldType in the case where the field is an enumeration.
        These 2 classes are mostly interchangeable.

        Args:
            project_id: the project to get the custom fields from

        Returns:
            list of all the custom fields

        References:
            testmanagement.getDefinedTestRunCustomFieldTypes
        """
        if not cls._custom_field_cache[project_id]:
            cfts = cls.session.test_management_client.service. \
                getDefinedTestRunCustomFieldTypes(project_id)
        else:
            cfts = cls._custom_field_cache[project_id]
        results = [CustomFieldType(suds_object=item)
                   if isinstance(item,
                                 CustomFieldType()._suds_object.__class__)
                   else EnumCustomFieldType(suds_object=item)
                   for item in cfts]
        return results

    def _cache_custom_fields(self, project_id):
        """Polarion API provides the custom fields of a TestRun.
        Get all custom fields and save in the dict self._custom_field_cache

        Args:
            project_id
        Returns
            None
        """
        self._custom_field_cache[project_id] = {}
        results = self.get_defined_custom_field_types(project_id)
        for result in results:
            f = getattr(result, "enum_id", None)
            key = result.cft_id
            f_type = None
            if f:
                f_type = self._custom_field_types(f)
            self._custom_field_cache[project_id][key] = {}
            self._custom_field_cache[project_id][key]["type"] = f_type
            self._custom_field_cache[project_id][key]["required"] = getattr(
                    result, "required", False)
            self._custom_field_cache[project_id][key]["multi"] = getattr(
                    result, "multi", False)

    def _add_custom_fields(self, project_id):
        """ This generates object attributes, with validation, so that custom
        fields can be related to as regular attributes. It takes the custom
        fields from the cache and calls the cache function if the values are
        not already there.
        Args:
            project_id - Each project can have its own custom fields. This
                function takes the custom fields for the specific project and
                adds them to the _cls_suds_map so they will be built into
                object attributes.
        Returns:
            None
        """
        self._changed_fields = {}
        # force the session to initialize. This is needed here because the
        # system has not yet been initialized.
        self.session
        if not project_id:
            project_id = self.default_project
        if project_id not in self._custom_field_cache:
            self._cache_custom_fields(project_id)
        cache = self._custom_field_cache[project_id]
        self._required_fields = []
        for field in cache:
            if cache[field]["required"]:
                self._required_fields.append(field)
            self._cls_suds_map[field] = {}
            self._cls_suds_map[field]["field_name"] = field
            self._cls_suds_map[field]["is_custom"] = True
            if cache[field]["type"] == Text:
                self._cls_suds_map[field]["cls"] = Text
            elif cache[field]["type"]:
                if cache[field]["multi"]:
                    self._cls_suds_map[field]["cls"] = ArrayOfEnumOptionId
                    self._cls_suds_map[field]["is_array"] = True
                else:
                    self._cls_suds_map[field]["cls"] = EnumOptionId
                self._cls_suds_map[field]["enum_id"] = cache[field]["type"]
                if isinstance(cache[field]["type"], type) and "project_id" in \
                        cache[field]["type"].__init__.__code__.co_varnames[
                        :cache[field]["type"].__init__.__code__.co_argcount]:
                    self._cls_suds_map[field]["additional_parms"] = \
                        {"project_id": project_id}

    def _get_index_of_test_record(self, test_case_id):
        # specific functions request the index of the test record within the
        # test run. However, the user doesn't know what the index is.
        # Therefore the user passes in the test case id and this function
        # figures out the index.
        # However, this function does not work for update_test_record_by_object
        # as that function requires the actual index of the record and not the
        # index of only executed records.
        index = -1
        for test_record in self._records:
            if test_record.executed:
                index += 1
            if test_case_id in test_record._suds_object.testCaseURI:
                return index
        raise PyleroLibException("The Test Case is either not part of "
                                 "this TestRun or has not been executed")

    def _status_change(self):
        # load a new object to test if the status should be changed.
        # can't use existing object because it doesn't include the new test rec
        # if the status needs changing, change it in the new object, so it
        # doesn't update any user made changes in the existing object.
        check_tr = TestRun(uri=self.uri)
        results = [rec.result for rec in check_tr.records if rec.result]
        if not results:
            status = "notrun"
            check_tr.finished_on = None
        elif len(results) == len(check_tr.records):
            status = "finished"
            # Do not touch finished_on because it can not be reverted
            # DPP-171495 Web services: You cannot reset finishedOn in TestRun
            # check_tr.finished_on = datetime.datetime.now()
        else:
            status = "inprogress"
            check_tr.finished_on = None
        if status != check_tr.status:
            check_tr.status = status
            check_tr.update()

    def _verify_record_count(self, record_index):
        # verifies the number of records is not less then the index given.
        self._verify_obj()
        if record_index > (len(self.records) - 1):
            raise PyleroLibException("There are only {0} test records".
                                     format(len(self.records)))

    def _verify_test_step_count(self, record_index, test_step_index):
        # verifies the number of test steps is not less then the index given.
        self._verify_record_count(record_index)
        if test_step_index > \
                (len(self.records[record_index].test_step_results) - 1):
            raise PyleroLibException("There are only {0} test records".
                                     format(len(self.records)))

    def add_attachment_to_test_record(self, test_case_id, path, title):
        """method add_attachment_to_test_record, adds the given attachment to
        the specified test record

        Args:
            test_case_id (str): The id of the test case
            path: file path to upload
            title: u.User friendly name of the file

        Returns:
            None

        Notes:
            Raises an error if the test case given is not in the TestRun or has
            not been executed yet.

        References:
            test_management.addAttachmentToTestRecord
        """
        record_index = self._get_index_of_test_record(test_case_id)
        self._verify_record_count(record_index)
        data = self._get_file_data(path)
        filename = os.path.basename(path)
        self.session.test_management_client.service. \
            addAttachmentToTestRecord(self.uri, record_index, filename,
                                      title, data)

    def add_attachment(self, path, title):
        """method add_attachment adds the given attachment to the current
        test run

        Args:
            path: file path to upload
            title: u.User friendly name of the file

        Returns:
            None

        Notes:
            Raises an error if the test run object is not populated

        References:
            test_management.addAttachmentToTestRun
        """
        self._verify_obj()
        data = self._get_file_data(path)
        filename = os.path.basename(path)
        self.session.test_management_client.service. \
            addAttachmentToTestRun(self.uri, filename, title, data)

    def add_attachment_to_test_step(self, test_case_id, test_step_index,
                                    path, title):
        """method add_attachment_to_test_step, adds the given attachment to
        the specified test step of the specified test record

        Args:
            test_case_id (str): The id of the test case to the step is in
            test_step_index (int): The 0 based index of the test step
            path: file path to upload
            title: u.User friendly name of the file

        Returns:
            none

        Notes:
            Raises an error if the record_index given is higher then the number
            of test records. or if the test_step_index is higher then the
            number of steps.

        References:
            test_management.addAttachmentToTestStep
        """
        record_index = self._get_index_of_test_record(test_case_id)
        self._verify_test_step_count(record_index, test_step_index)
        data = self._get_file_data(path)
        filename = os.path.basename(path)
        self.session.test_management_client.service.addAttachmentToTestStep(
            self.uri, record_index, test_step_index, filename, title, data)

    def _check_test_record_exists(self, test_case_id):
        """Searches the test run to see if the case is already a member of it
        and if so raises an exception. It should only receive one result set
        and if it returns more then that it raises a different exception.
        It searches the server for the test record in case, the record was
        added by another process.

        Notes:
            This is the best that can be done and there is no method of knowing
            if another process added the same record in between this check and
            the actual add.

        Args:
            test_case_id (str): the id of the test case to check

        Returns:
            None
        """
        check_tr = TestRun.search(
            'project.id:%s AND id:"%s" AND %s' %
            (self.project_id, self.test_run_id, test_case_id))
        if len(check_tr) not in [0, 1]:
            raise PyleroLibException(
                "The search function did not work as expected. Please report.")
        elif len(check_tr) == 1:
            raise PyleroLibException(
                "This test case is already part of the test run")
        else:
            return None

    @tx_wrapper
    def add_test_record_by_fields(self, test_case_id, test_result,
                                  test_comment, executed_by, executed,
                                  duration, defect_work_item_id=None):
        """method add_test_record_by_fields, adds a test record for the given
        test case based on the result fields passed in.
        When a test record is added, it changes the test run status to
        "inprogress" and when the last test record is run, it changes the
        status to done.

        Args:
            test_case_id (str): The id of the test case that was executed
            test_result (str): Must be one of the following values:
                                  passed
                                  failed
                                  blocked
            test_comment (str or Text object): may be None
            executed_by (str): user id
            executed (datetime):
            duration (float):
            defect_work_item_id (str): _WorkItem id of defect, default: None

        Returns:
            None

        References:
            test_management.addTestRecordToTestRun
        """
        self._verify_obj()
        self.check_valid_field_values(test_result, "result", {})
        if not executed or not test_result:
            raise PyleroLibException(
                "executed and test_result require values")
        testrec = TestRecord(self.project_id, test_case_id)
        testrec.result = test_result
        testrec.comment = test_comment
        testrec.executed_by = executed_by
        testrec.executed = executed
        testrec.duration = duration
        if defect_work_item_id:
            testrec.defect_case_id = defect_work_item_id
        self.add_test_record_by_object(testrec)

    @tx_wrapper
    def add_test_record_by_object(self,
                                  test_record,
                                  manual_state_change=False):
        """method add_test_record_by_object, adds a test record for the given
        test case based on the TestRecord object passed in.
        In addition, the test run is checked for completeness and the test
        run state will change accordingly.
        Test Run states are ["notrun", "finished", "inprogress"].

        Args:
            test_record (TestRecord or Polarion TestRecord):
            manual_state_change (boolean): Change test run state
                                           automatically or manual.

        Returns:
            None

        References:
            test_management.addTestRecordToTestRun
        """
        self._verify_obj()
        test_case_id = test_record.test_case_id
        self._check_test_record_exists(test_case_id)
        if isinstance(test_record, TestRecord):
            suds_object = test_record._suds_object
        elif isinstance(test_record, TestRecord()._suds_object.__class__):
            suds_object = test_record
        if test_record.result == "failed" and not test_record.defect_case_id:
            test_record.defect_case_id = \
                create_incident_report(self, test_record,
                                       TestCase(work_item_id=test_case_id))
        self.session.test_management_client.service.addTestRecordToTestRun(
            self.uri, suds_object)

        if manual_state_change:
            self.update()
        else:
            self._status_change()

    def create_summary_defect(self, defect_template_id=None):
        """method create_summary_defect, adds a new summary _WorkItem for the
        test case based on the _WorkItem template id passed in. If not template
        is passed in, it creates it based on the default template.

        Args:
            defect_template_id (str): the _WorkItem template id to base the
            new summary defect. can be null. default: None

        Returns:
            the created _WorkItem

        References:
            test_management.createSummaryDefect
        """

        self._verify_obj()
        defect_template_uri = None
        if defect_template_id:
            suds_defect_template = _WorkItem(work_item_id=defect_template_id,
                                             project_id=self.project_id)
            defect_template_uri = suds_defect_template._uri
        wi_uri = self.session.test_management_client.service. \
            createSummaryDefect(self.uri, defect_template_uri)
        return _WorkItem(uri=wi_uri)

    def delete_attachment_from_test_record(self, test_case_id, filename):
        """Deletes Test Record Attachment of specified record and
        attachment's file name.

        Args:
            test_case_id: The test case to delete the attachment from
            filename: name of the file to delete

        Returns:
            None

        References:
            test_management.deleteAttachmentFromTestRecord
        """
        record_index = self._get_index_of_test_record(test_case_id)
        self._verify_record_count(record_index)
        self.session.test_management_client.service. \
            deleteAttachmentFromTestRecord(self.uri, record_index, filename)

    def delete_attachment_from_test_step(self, test_case_id, test_step_index,
                                         filename):
        """Deletes Test Step Attachment of the specified step in the specified
        test record.

        Args:
            test_case_id: The test case to delete the attachment from
            filename: name of the file to delete

        Returns:
            None

        References:
            test_management.deleteAttachmentFromTestRecord
        """
        record_index = self._get_index_of_test_record(test_case_id)
        self._verify_test_step_count(record_index, test_step_index)
        self.session.test_management_client.service. \
            deleteAttachmentFromTestStep(self.uri, record_index,
                                         test_step_index, filename)

    def delete_attachment(self, filename):
        """Deletes Test Run Attachment specified by attachment's
        file name. Method is applicable also on Test Run Template.

        Args:
            filename: filename to delete

        Returns:
            None

        References:
            test_management.deleteTestRunAttachment
        """
        self._verify_obj()
        self.session.test_management_client.service. \
            deleteTestRunAttachment(self.uri, filename)

    def get_attachment(self, filename):
        """Gets Test Run Attachment specified by attachment's
        file name. Method is applicable also on Test Run Template.

        Args:
            filename: filename to delete

        Returns:
            TestRunAttachment object

        References:
            test_management.getTestRunAttachment
        """
        self._verify_obj()
        suds_attach = self.session.test_management_client.service. \
            getTestRunAttachment(self.uri, filename)
        return TestRunAttachment(suds_object=suds_attach)

    def get_attachments(self):
        """method get_attachments returns all the attachments for the TestRun

        Args:
            None

        Returns:
            ArrayOfTestRunAttachments object

        References:
            test_management.getTestRunAttachments
        """
        self._verify_obj()
        lst_suds_attach = self.session.test_management_client.service. \
            getTestRunAttachments(self.uri)
        lst_attach = [TestRunAttachment(suds_object=suds_attach)
                      for suds_attach in lst_suds_attach]
        return lst_attach

    def get_custom_field(self, field_name):
        """gets custom field values.

        Args:
            field_name: name of the custom field

        Returns:
            value of the custom field.

        Note: Polarion WSDL currently does not publish the list of custom
              fields, so this function cannot do any verification if the field
              is valid.
        """
        self._verify_obj()
        cf = self._custom_fields
        match = [x for x in cf if x.key == field_name]
        if match:
            return match[0].value
        else:
            return Custom(field_name, None)

    def get_wiki_content(self):
        """method get_wiki_content returns the wiki content for the Test Run

        Args:
            None

        Returns:
            Text object containing the wiki content

        References:
            test_management.getWikiContentForTestRun
        """
        self._verify_obj()
        suds_wiki = self.session.test_management_client.service. \
            getWikiContentForTestRun(self.uri)
        return Text(suds_object=suds_wiki)

    def _set_custom_field(self, field_name, value):
        """sets custom field values.

        Args:
            field_name: name of the custom field

        Returns:
            value of the custom field.

        Note: Polarion WSDL currently does not publish the list of custom
              fields, so this function cannot do any verification if the field
              or value is valid.
        """
        self._verify_obj()
        cf = self._custom_fields
        cust = Custom()
        cust.key = field_name
        cust.value = value
        if cf:
            # check if the custom field already exists and if so, modify it.
            match = [x for x in cf if x.key == field_name]
            if match:
                match[0].value = value
            else:
                cf.append(cust)
        else:
            cf = [cust]
        self._custom_fields = cf

    def update(self):
        """method update updates the testRun object with the attribute values
        currently in the object.

        Args:
            None

        Returns
            None

        References:
            test_management.updateTestRun
        """
        self._verify_obj()
        self.verify_required()
        for field in self._changed_fields:
            self._set_custom_field(field, self._changed_fields[field])
        self._changed_fields = {}
        self.session.test_management_client.service.updateTestRun(
            self._suds_object)

    def update_attachment(self, path, original_filename, title):
        """method update_attachment updates the specified attachment to the
        current test run

        Args:
            path: file path to upload
            original_filename: The file that we want to overwrite with the new
                               file
            title: u.User friendly name of the file

        Returns:
            None

        Notes:
            Raises an error if the test run object is not populated.

        References:
            test_management.updateTestRunAttachment
        """
        self._verify_obj()
        data = self._get_file_data(path)
        filename = os.path.basename(original_filename)
        self.session.test_management_client.service. \
            updateTestRunAttachment(self.uri, filename, title, data)

    def update_summary_defect(self, source, total_failures, total_errors,
                              total_tests, defect_template_id):
        """method update-summary_defect creates or updates the summary defect
        Work Item of a test run.

        Args:
            source (string): source of the summary defect, used to generate
                              the description content.
            total_failures (int): amount of total failures in the test run,
                                  used to generate the description content.
            total_errors (int): amount of total errors in the test run,
                                used to generate the description content.
            total_tests (int): amount of total tests in the test run,
                               used to generate the description content.
            defect_template_id: ID of the defect template Work Item to be
                                used, the configured template will be used if
                                None.

        Returns:
            the created or updated _WorkItem

        References:
            test_management.updateSummaryDefect
        """
        self._verify_obj()
        suds_defect_template = _WorkItem(work_item_id=defect_template_id,
                                         project_id=self.project_id)
        defect_template_uri = suds_defect_template._uri
        wi_uri = self.session.test_management_client.service. \
            updateSummaryDefect(self.uri, source, total_failures, total_errors,
                                total_tests, defect_template_uri)
        return _WorkItem(uri=wi_uri)

    @tx_wrapper
    def update_test_record_by_fields(self, test_case_id,
                                     test_result,
                                     test_comment,
                                     executed_by,
                                     executed,
                                     duration,
                                     defect_work_item_id=None):
        """method update_test_record_by_fields updates a test record.

        Args:
            test_case_id: id of the test case to update.
            test_result: Must be one of the following values:
                                   passed
                                   failed
                                   blocked
            test_comment: (str or Text object) - may be None
            executed_by (str): user id
            executed: date when the test case has been executed
            duration: duration of the test case execution, any negative value
                      is treated as None.
            defect_work_item_id: _WorkItem id of defect, can be None
                                Default: None

        Returns:
            None

        Notes:
            Only a test case that has already been executed may be updated
            using this function. To execute a test record use the
            add_test_record_by_fields function

        References:
            test_management.updateTestRecord
        """
        self._verify_obj()
        testrec = TestRecord(self.project_id, test_case_id)
        testrec.result = test_result
        testrec.comment = test_comment
        testrec.executed_by = executed_by
        testrec.executed = executed
        testrec.duration = duration
        if defect_work_item_id:
            testrec.defect_case_id = defect_work_item_id
        self.update_test_record_by_object(test_case_id, testrec)

    @tx_wrapper
    def update_test_record_by_object(self,
                                     test_case_id,
                                     test_record,
                                     manual_state_change=False):
        """method update_test_record_by_object, adds a test record for the
        given test case based on the TestRecord object passed in.
        In addition, the test run is checked for completeness and the
        test run state will change accordingly.
        Test Run states are ["notrun", "finished", "inprogress"].

        Args:
            test_case_id (str): the test case id that the record is related to.
            test_record (TestRecord or Polarion TestRecord)
            manual_state_change (boolean): Change the test
                                           run result automatically or manual.

        Returns:
            None

        References:
            test_management.updateTestRecordAtIndex
        """
        self._verify_obj()
        # this function cannot use the _get_index_of_test_record function
        # because this function (specifically and not documented) uses the
        # actual index of the test records and not the index of all
        # executed records.
        test_case_ids = [rec.test_case_id for rec in self._records]
        if test_case_id not in test_case_ids:
            self.add_test_record_by_object(test_record)
        else:
            if test_record.result == "failed" and \
                    not test_record.defect_case_id:
                test_record.defect_case_id = \
                    create_incident_report(self, test_record,
                                           TestCase(work_item_id=test_case_id))
            index = test_case_ids.index(test_case_id)
            if isinstance(test_record, TestRecord):
                suds_object = test_record._suds_object
            elif isinstance(test_record, TestRecord()._suds_object.__class__):
                suds_object = test_record
            self.session.test_management_client.service. \
                updateTestRecordAtIndex(self.uri, index, suds_object)
            if manual_state_change:
                self.update()
            else:
                self._status_change()

    def update_wiki_content(self, content):
        """method update_wiki_content updates the wiki for the current TestRun

        Args:
            Content (str or Text object)

        Returns:
            None

        References:
            test_management.updateWikiContentForTestRun
        """
        self._verify_obj()
        if content:
            if isinstance(content, basestring):
                obj_content = Text(content=content)
                suds_content = obj_content._suds_object
            elif isinstance(content, Text):
                suds_content = content._suds_object
            else:  # is a suds object
                suds_content = content
        else:
            suds_content = suds.null()
        self.session.test_management_client.service. \
            updateWikiContentForTestRun(self.uri, suds_content)

    def verify_required(self):
        """function that checks if all required fields are passed in

        Exceptions:
            PyleroLibException - if required params are not passed in
        """
        fields = ""
        for req in self._required_fields:
            if not getattr(self, req):
                fields += (", " if fields else "") + req
        if fields:
            raise PyleroLibException("These parameters are required: {0}".
                                     format(fields))

    def verify_params(self, **kwargs):
        """function that checks if all the kwargs are valid attributes.

        Args:
            **kwargs: keyword arguments. Test run attributes can be passed in
                      to be set upon creation. Required fields must be passed
                      in
        Exceptions:
            PyleroLibException - if params are unknown
        """
        params = ""
        for param in kwargs:
            if param not in self._cls_suds_map:
                params += (", " if params else "") + param

        if params:
            raise PyleroLibException("These parameters are unknown: {0}".
                                     format(params))



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.properties import Properties


class EnumOption(BasePolarion):
    """Object to handle the Polarion WSDL tns5:EnumOption class

    Attributes:
        default (boolean)
        enum_id (string)
        hidden (boolean)
        enum_option_id (string)
        name (string)
        phantom (boolean)
        properties (Properties)
        sequence_number (int)
"""
    _cls_suds_map = {"default": "default",
                     "enum_id": "enumId",
                     "hidden": "hidden",
                     "enum_option_id": "id",
                     "name": "name",
                     "phantom": "phantom",
                     "properties":
                     {"field_name": "properties",
                      "cls": Properties},
                     "sequence_number": "sequenceNumber",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns5:EnumOption"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class ActivitySource(BasePolarion):
    """Object to handle the Polarion WSDL tns3:ActivitySource class

    Attributes:
        activity_source_id (string)
        prefix (string)
        types (ArrayOf_xsd_string)
"""
    _cls_suds_map = {"activity_source_id": "id",
                     "prefix": "prefix",
                     "types": "types",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "tracker_client"
    _obj_struct = "tns3:ActivitySource"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class Property(BasePolarion):
    """Object to handle the Polarion WSDL tns4:property class

    Attributes:
        key (string)
        value (string)
"""
    _cls_suds_map = {"key": "key",
                     "value": "value",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns4:property"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class CustomField(BasePolarion):
    """Object to handle the Polarion WSDL tns3:CustomField class

    Attributes:
        key (string)
        parent_item_uri (string)
        value (anyType)
"""
    _cls_suds_map = {"key": "key",
                     "parent_item_uri": "parentItemURI",
                     "value": "value"}
    _obj_client = "tracker_client"
    _obj_struct = "tns3:CustomField"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.enum_option_id import ArrayOfEnumOptionId
from pylero.enum_option_id import EnumOptionId
from pylero.imported_comment import ImportedComment
from pylero.signature_data import SignatureData
from pylero.subterra_uri import ArrayOfSubterraURI
from pylero.subterra_uri import SubterraURI
from pylero.text import Text
from pylero.user import User


class ModuleComment(BasePolarion):
    """Object to handle the Polarion WSDL tns4:ModuleComment class

    Attributes:
        author (User)
        child_comment_uris (ArrayOfSubterraURI)
        created (dateTime)
        module_comment_id (string)
        imported_comment (ImportedComment)
        parent_comment_uri (SubterraURI)
        referred_work_item_uri (SubterraURI)
        resolved (boolean)
        signature_data (SignatureData)
        tags (ArrayOfEnumOptionId)
        text (Text)
"""
    _cls_suds_map = {"author":
                     {"field_name": "author",
                      "cls": User},
                     "child_comment_uris":
                     {"field_name": "childCommentURIs",
                      "is_array": True,
                      "cls": SubterraURI,
                      "arr_cls": ArrayOfSubterraURI,
                      "inner_field_name": "SubterraURI"},
                     "created": "created",
                     "module_comment_id": "id",
                     "imported_comment":
                     {"field_name": "importedComment",
                      "cls": ImportedComment},
                     "parent_comment_uri":
                     {"field_name": "parentCommentURI",
                      "cls": SubterraURI},
                     "referred_work_item_uri":
                     {"field_name": "referredWorkItemURI",
                      "cls": SubterraURI},
                     "resolved": "resolved",
                     "signature_data":
                     {"field_name": "signatureData",
                      "cls": SignatureData},
                     "tags":
                     {"field_name": "tags",
                      "is_array": True,
                      "cls": EnumOptionId,
                      "arr_cls": ArrayOfEnumOptionId,
                      "inner_field_name": "EnumOptionId"},
                     "text":
                     {"field_name": "text",
                      "cls": Text},
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "test_management_client"
    _obj_struct = "tns4:ModuleComment"


class ArrayOfModuleComment(BasePolarion):
    _obj_client = "test_management_client"
    _obj_struct = "tns4:ArrayOfModuleComment"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.build_linked_work_item import ArrayOfBuildLinkedWorkItem
from pylero.build_linked_work_item import BuildLinkedWorkItem
from pylero.build_test_results import BuildTestResults
from pylero.user import User


class Build(BasePolarion):
    """Object to handle the Polarion WSDL tns2:Build class

    Attributes:
        author (User)
        bir_location (Location)
        build_descriptor_name (string)
        build_stamp (string)
        build_status (string)
        build_tag (string)
        build_test_results (BuildTestResults)
        calculation_descriptor_name (string)
        creation_time (dateTime)
        finish_time (dateTime)
        build_id (string)
        job_id (string)
        linked_work_items (ArrayOfBuildLinkedWorkItem)
        local_deployment_space_name (string)
        log_files (ArrayOfLocation)
        start_time (dateTime)
"""
    _cls_suds_map = {"author":
                     {"field_name": "author",
                      "cls": User},
                     "bir_location": "birLocation",
                     "build_descriptor_name": "buildDescriptorName",
                     "build_stamp": "buildStamp",
                     "build_status": "buildStatus",
                     "build_tag": "buildTag",
                     "build_test_results":
                     {"field_name": "buildTestResults",
                      "cls": BuildTestResults},
                     "calculation_descriptor_name":
                     "calculationDescriptorName",
                     "creation_time": "creationTime",
                     "finish_time": "finishTime",
                     "build_id": "id",
                     "job_id": "jobId",
                     "linked_work_items":
                     {"field_name": "linkedWorkItems",
                      "is_array": True,
                      "cls": BuildLinkedWorkItem,
                      "arr_cls": ArrayOfBuildLinkedWorkItem,
                      "inner_field_name": "BuildLinkedWorkItem"},
                     "local_deployment_space_name": "localDeploymentSpaceName",
                     "log_files": "logFiles",
                     "start_time": "startTime",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns2:Build"



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


# Unknown what this class object is supposed to do. Will determine during
# more intensive testing
class SubterraURI(BasePolarion):
    def __init__(self):
        pass


class ArrayOfSubterraURI(BasePolarion):
    def __init__(self):
        pass



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.project import Project
from pylero.subterra_uri import ArrayOfSubterraURI
from pylero.subterra_uri import SubterraURI
from pylero.text import Text
from pylero.user import User
from pylero.wiki_page_attachment import ArrayOfWikiPageAttachment
from pylero.wiki_page_attachment import WikiPageAttachment


class WikiPage(BasePolarion):
    """Object to handle the Polarion WSDL tns3:WikiPage class

    Attributes:
        attachments (ArrayOfWikiPageAttachment)
        author (User)
        created (dateTime)
        home_page_content (Text)
        wiki_page_id (string)
        linked_page_uris (ArrayOfSubterraURI)
        location (Location)
        page_location (Location)
        page_name (string)
        project (Project)
        space_id (string)
        title (string)
        type (string)
        updated (dateTime)
        updated_by (User)
"""
    _cls_suds_map = {"attachments":
                     {"field_name": "attachments",
                      "is_array": True,
                      "cls": WikiPageAttachment,
                      "arr_cls": ArrayOfWikiPageAttachment,
                      "inner_field_name": "WikiPageAttachment"},
                     "author":
                     {"field_name": "author",
                      "cls": User},
                     "created": "created",
                     "home_page_content":
                     {"field_name": "homePageContent",
                      "cls": Text},
                     "wiki_page_id": "id",
                     "linked_page_uris":
                     {"field_name": "linkedPageURIs",
                      "is_array": True,
                      "cls": SubterraURI,
                      "arr_cls": ArrayOfSubterraURI,
                      "inner_field_name": "SubterraURI"},
                     "location": "location",
                     "page_location": "pageLocation",
                     "page_name": "pageName",
                     "project":
                     {"field_name": "project",
                      "cls": Project},
                     "space_id": "spaceId",
                     "title": "title",
                     "type": "type",
                     "updated": "updated",
                     "updated_by":
                     {"field_name": "updatedBy",
                      "cls": User},
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "tracker_client"
    _obj_struct = "tns3:WikiPage"

    @classmethod
    def get_wiki_pages(cls, project_id, space_id, fields):
        """Returns Wiki Pages from given project and space.

        Args:
            project_id: project id (not null)
            space_id: space id (not null)

        Returns:
            list of WikiPage objects
        """
        function_name = "getWikiPages"
        parms = [project_id, space_id]
        if fields:
            function_name += "WithFields"
            parms += [cls._convert_obj_fields_to_polarion(fields)]
        wikis = []
        for suds_wiki in getattr(cls.session.tracker_client.service,
                                 function_name)(*parms):
            wikis.append(cls(suds_object=suds_wiki))
        return wikis

    @classmethod
    def query(cls, query, is_sql=False, fields=["wiki_page_id"],
              sort="wiki_page_id", limit=-1, baseline_revision=None,
              query_uris=False):
        """Searches for Wiki Pages .

        Args:
            query: query, either Lucene or SQL
            is_sql (bool): determines if the query is SQL or Lucene
            fields: list of field names to fill in the returned
                     Modules/Documents (can be null). For nested structures in
                     the lists you can use following syntax to include only
                     subset of fields: myList.LIST.key
                     (e.g. linkedWorkItems.LIST.role).
                     For custom fields you can specify which fields you want to
                     be filled using following syntax:
                     customFields.CUSTOM_FIELD_ID (e.g. customFields.risk).
                     default - list containing "wiki_page_id"
            sort: Lucene sort string, default wiki_page_id
            limit: how many results to return (-1 means everything (default))
            baseline_revision (str): if populated, query done in specified rev
                                     default - None
            query_uris: returns a list of URI of the Modules found, instead of
                        a list of WikiPage objects. default - False

        Returns:
            list of modules

        References:
            queryWikiPageUris
            queryWikiPageUrisBySQL
            queryWikiPageUrisInBaseline
            queryWikiPageUrisInBaselineBySQL
            queryWikiPages
            queryWikiPagesBySQL
            queryWikiPagesInBaseline
            queryWikiPagesInBaselineBySQL
        """
        parms = [query]
        # The parameters have to be listed in the specific order, based on the
        # specific function called. That's why there are 2 if not is_sql
        # conditions.
        if not is_sql:
            parms.append(sort)
        if baseline_revision:
            parms.append(baseline_revision)
        if not query_uris:
            p_fields = cls._convert_obj_fields_to_polarion(fields)
            parms.append(p_fields)
        if not is_sql:
            parms.append(limit)
        if not query_uris:
            base_name = "queryWikiPages"
        else:
            base_name = "queryWikiPageUris"
        if baseline_revision:
            base_name += "InBaseline"
        if is_sql:
            base_name += "BySQL"
        wps = getattr(cls.session.tracker_client.service, base_name)(*parms)
        if query_uris:
            return wps
        else:
            lst_wp = [WikiPage(suds_object=wp) for wp in wps]
            return lst_wp

    def __init__(self, fields=None, uri=None, suds_object=None):
        """
            Args:
                fields: list of object fields to be returned in the object

            Returns:
                None

            References:
                Tracker.getWikiPageByUri
                tracker.getWikiPageByUriWithFields
        """
        super(self.__class__, self).__init__(suds_object=suds_object)
        if uri:
            function_name = "getWikiPageByUri"
            parms = [uri]
            if fields:
                function_name += "WithFields"
                parms += [self._convert_obj_fields_to_polarion(fields)]
            self._suds_object = getattr(self.session.tracker_client.service,
                                        function_name)(*parms)



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.signature import ArrayOfSignature
from pylero.signature import Signature
from pylero.user import User


class SignatureContext(BasePolarion):
    """Object to handle the Polarion WSDL tns4:SignatureContext class

    Attributes:
        signatures (ArrayOfSignature)
        target_status_id (string)
        transition_data_revision (string)
        user (User)
"""
    _cls_suds_map = {"transition_data_revision": "transitionDataRevision",
                     "signatures":
                     {"field_name": "signatures",
                      "is_array": True,
                      "cls": Signature,
                      "arr_cls": ArrayOfSignature,
                      "inner_field_name": "Signature"},
                     "target_status_id": "targetStatusId",
                     "user":
                     {"field_name": "user",
                      "cls": User}}
    _obj_client = "test_management_client"
    _obj_struct = "tns4:SignatureContext"


class ArrayOfSignatureContext(BasePolarion):
    _obj_client = "test_management_client"
    _obj_struct = "tns4:ArrayOfSignatureContext"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import base64
import copy
import os
import re
from functools import wraps
from getpass import getpass

import suds
from pylero._compatible import basestring
from pylero._compatible import classmethod
from pylero._compatible import ConfigParser
from pylero.exceptions import PyleroLibException
from pylero.server import Server


# classproperty is a property that works on the class level
class ClassProperty(property):
    """Returns a classmethod as the getter so that the property can be used as
    a class property. This is needed so that the property can be set for all
    child objects. This project currently has no need of a setter or deleter.
    """

    def __get__(self, instance, cls):
        return classmethod(self.fget).__get__(instance, cls)()


class Configuration(object):
    pkgdir = os.path.dirname(__file__)
    GLOBAL_CONFIG = "%s/etc/pylero.cfg" % pkgdir
    LOCAL_CONFIG = os.path.expanduser("~") + "/.pylero"
    CURDIR_CONFIG = ".pylero"
    CONFIG_SECTION = "webservice"
    # Look at ConfigParser
    # https://docs.python.org/2.6/library/configparser.html

    def __init__(self):
        defaults = {"cachingpolicy": "0",
                    "timeout": "120"}
        config = ConfigParser(defaults)
        # Check for existence of config file and config_section
        if not config.read([self.GLOBAL_CONFIG, self.LOCAL_CONFIG,
                            self.CURDIR_CONFIG]) or \
                not config.has_section(self.CONFIG_SECTION):
            # Check for mandatory environ variables if config file is not found
            if not all(os.environ.get(item) for item in ('POLARION_URL',
                                                         'POLARION_REPO',
                                                         'POLARION_USERNAME',
                                                         'POLARION_PASSWORD',
                                                         'POLARION_PROJECT')):
                raise PyleroLibException("The config files/ENV vars do not "
                                         "exist or are not of the correct "
                                         "format. Valid files are: {0}, {1} "
                                         "or {2}" .format(self.GLOBAL_CONFIG,
                                                          self.LOCAL_CONFIG,
                                                          self.CURDIR_CONFIG))
        self.server_url = os.environ.get("POLARION_URL") or \
            config.get(self.CONFIG_SECTION, "url")
        self.repo = os.environ.get("POLARION_REPO") or \
            config.get(self.CONFIG_SECTION, "svn_repo")
        self.login = os.environ.get("POLARION_USERNAME") or \
            config.get(self.CONFIG_SECTION, "user")
        self.pwd = os.environ.get("POLARION_PASSWORD") or \
            config.get(self.CONFIG_SECTION, "password")

        try:
            self.timeout = os.environ.get("POLARION_TIMEOUT") or \
                config.get(self.CONFIG_SECTION, "timeout")
        except Exception:
            self.timeout = config.defaults['timeout']

        try:
            self.timeout = int(self.timeout)
        except ValueError:
            raise PyleroLibException("The timeout value in the config"
                                     " file must be an integer")
        self.proj = os.environ.get("POLARION_PROJECT") or \
            config.get(self.CONFIG_SECTION, "default_project")
        try:
            self.cert_path = os.environ.get("POLARION_CERT_PATH") or \
                config.get(self.CONFIG_SECTION, "cert_path")
        except Exception:
            self.cert_path = None

        if not (self.server_url and self.login and self.proj):
            raise PyleroLibException("The config files must contain "
                                     "valid values for: url, user, "
                                     "password and default_project")

        try:
            self.disable_manual_auth = \
                os.environ.get("POLARION_DISABLE_MANUAL_AUTH") or \
                config.getboolean(self.CONFIG_SECTION,
                                  "disable_manual_auth")
        except Exception:
            self.disable_manual_auth = False


class Connection(object):
    """Creates a Polarion session as a class method, so that it is used for all
    objects inherited by BasePolarion.
    The url, repo, user and password are read from config files, which are
    located either the current directory ./pylero, the user's dir ~/.pylero
    or the Library config dir LIBDIR/etc/pylero.cfg
    These can also be overridden with the following environment variables:
    POLARION_URL
    POLARION_REPO
    POLARION_USERNAME
    POLARION_PASSWORD
    POLARION_TIMEOUT
    POLARION_PROJECT
    """
    connected = False
    session = None
    password_retries = 3

    @classmethod
    def session(cls):
        if not cls.connected:
            cfg = Configuration()
            # if the password is not supplied in the config file, ask the user
            # for it
            if not cfg.pwd:
                cfg.pwd = getpass(
                    "Password not in config file.\nEnter Password:")
            while not cls.connected and cls.password_retries:
                try:
                    srv = Server(
                        cfg.server_url,
                        cfg.login, cfg.pwd,
                        timeout=cfg.timeout,
                        cert_path=cfg.cert_path)
                    cls.session = srv.session()
                    cls.session._login()
                    cls.connected = True
                except suds.WebFault as e:
                    # If we couldn't connect its because the user has typed the
                    # wrong password. So we keep asking for password till we
                    # are successfully connected
                    if "com.polarion.platform.security." \
                            "AuthenticationFailedException" \
                            in e.fault.faultstring:
                        if cfg.disable_manual_auth:
                            raise PyleroLibException("Manual authentication "
                                                     "is disabled")
                        cfg.pwd = getpass("Invalid Password.\nEnter Password:")
                        cls.password_retries -= 1
                    else:
                        raise
            if not cls.password_retries:
                raise PyleroLibException("Unable to establish pylero session "
                                         "due to 3 incorrect login attempts")
            cls.session.default_project = cfg.proj
            cls.session.user_id = cfg.login
            cls.session.password = cfg.pwd
            cls.session.repo = cfg.repo
            # must use try/except instead of or because the config file
            # may return a non empty value, such as " "
        return cls.session


def tx_wrapper(func):
    # decorator function to run specific functions in.
    # Because they have multiple modifying stmts, they should be run
    # in a transaction. It first checks if it is already in a tx and if
    # not, it starts one itself.
    @wraps(func)
    def inner(*args, **kwargs):
        # the first object is the instance or class object.
        self = args[0]
        new_tx = False
        try:
            if not self.session.tx_in():
                self.session.tx_begin()
                new_tx = True
            res = func(*args, **kwargs)
            if new_tx:
                self.session.tx_commit()
            return res
        except (suds.WebFault, PyleroLibException, Exception):
            if new_tx and self.session.tx_in():
                self.session.tx_rollback()
            raise
    return inner


class BasePolarion(object):
    """BasePolarion is the parent class for all the WSDL Polarion objects that
    are published. Using the _cls_suds_map, the class creates a property for
    each attribute so that any access of the object attribute, will access the
    WSDL object that is contained by it.

    Attributes:
        _cls_suds_map (dict): maps the Polarion attribute names to the Pylero
                        attribute names. Attributes that reference either
                        objects or an array of objects have the properties
                        relate to the relationship meaning that accessing the
                        property will give access to the object or list of
                        objects.
        _id_field (str): the field that represents an id field, used in the
                         child class's constructor. when a child's class
                         defines the field it allows this constructor to accept
                         an obj_id as a parameter
        _obj_client (str): The Polarion client child's WSDL object is
                           defined by
        _obj_struct (str): The data type defined by the WSDL library. The
                           structure of the datatype is tnsX:ObjectName, the X
                           is per datatype and has no default.
        session (Session): The Polarion Session object, initialized by the
                           Connection class. This attribute connects to the
                           server one time per session, no matter how many
                           objects are instantiated.
        default_project (str): The user's default project, to be used when
                          project_id is needed and there is none given
    """
    _cls_suds_map = {}
    _id_field = None
    _obj_client = None
    _obj_struct = None
    _session = None
    _default_project = None
    _cache = {
        "enums": {},
        "custom_field_types": {},
        "projects": {}
    }
    REGEX_PROJ = r"/default/(.*)\$"
    # The id in the uri is always after the last }, at times there are multiple
    REGEX_ID = ".+}(.*)$"
    # The URI_STRUCT can be overridden in a child class when needed (for
    # example, Documents). Also if there is need for a replace in the child
    # class the URI_ID lambda attributes should be overridden
    URI_STRUCT = "subterra:data-service:objects:/default/" \
                 "%(project)s${%(obj)s}%(id)s"
    # must wrap lambda with classmethod so it can be used as such
    URI_ID_GET_REPLACE = classmethod(lambda cls, x: x)
    URI_ID_SET_REPLACE = classmethod(lambda cls, x: x)

    @ClassProperty
    def session(cls):
        # Uses a class property for the session, so that the library doesn't
        # connect to the server until the library is actually used.
        if BasePolarion._session:
            return BasePolarion._session
        else:
            # For some reason, using the cls attribute makes it into a class
            # attribute for the specific class but not for all the other
            # Pylero objects.
            BasePolarion._session = Connection.session()
            BasePolarion._default_project = cls._session.default_project
            BasePolarion.logged_in_user_id = cls._session.user_id
            # stores password in the session so it can be used for direct svn
            # operations
            BasePolarion.repo = cls._session.repo
            return BasePolarion._session

    @ClassProperty
    def default_project(cls):
        # Uses a class property for the session, so that the library connects
        # when accessing it.
        if not BasePolarion._default_project:
            cls.session
        return BasePolarion._default_project

    @classmethod
    def _convert_obj_fields_to_polarion(cls, fields=[]):
        """All child methods that take a fields parameter should pass the field
        array and it converts it to the Polarion attribute name using the
        _cls_suds_map. It uses the python map function over the fields list to
        return the list of its Polarion attribute name. If the field is a
        custom field, it adds the customFields. qualifier before the fieldname
        as is required for searching custom fields.

        Args:
            fields - list of fields to convert. If it is not a list, it
            converts it to one first. default: []
        """
        p_fields = []
        if fields:
            if not isinstance(fields, list):
                fields = [fields]
            # convert given fields to Polarion fields
            p_fields = ["%s%s" % (
                "customFields."
                if isinstance(cls._cls_suds_map[x], dict) and
                cls._cls_suds_map[x].get("is_custom", False)
                else "",
                cls._cls_suds_map[x]
                if not isinstance(cls._cls_suds_map[x], dict)
                else cls._cls_suds_map[x]["field_name"]) for x in fields]
            # Omit 'URIs' and 'URI' from URIFields
            p_fields = [(x.replace("URIs", "").replace("URI", ""))
                        for x in p_fields]
        return p_fields

    @classmethod
    def get_global_roles(cls):
        r"""Returns all global roles.

        Args:
            None

        Returns:
            list of global roles

        References:
            Security.\ :security:`getGlobalRoles()`
        """
        return cls.session.security_client.service.getGlobalRoles()

    @classmethod
    def has_current_user_permission(cls, permission, project_id):
        """Checks if given permission is granted to the current user.

        Args:
            permission: the permission to check.
            project_id: the id of the project to check the permission in,
                        None to check global permissions.

        Returns:
            bool

        References:
            Security.hasCurrentUserPermission
        """
        return cls.session.security_client.service.hasCurrentuserPermission(
            permission, project_id)

    def __init__(self, obj_id=None, suds_object=None):
        # cls_suds_map must be available for some parameters on the class
        # level, but gets changed on the instance level and those changes
        # should not be accessible to other instances. This is the reason
        # for overwriting it as an instance attribute.
        self._cls_suds_map = copy.deepcopy(self._cls_suds_map)
        # _fix_circular_refs is a function that allows objects to contain
        # circular references by applying the reference only after the class
        # has been instantiated. Some objects contain references to themselves,
        # for example a parent attribute.
        if hasattr(self, "_fix_circular_refs"):
            self._fix_circular_refs()
        # if _id_field has not been set in the child class, the obj_id field
        # cannot be passed as a parameter.
        if obj_id and not self._id_field:
            raise PyleroLibException(
                "{0} only accepts a suds object, not an obj_id".format(
                    self.__class__.__name))
        if suds_object:
            self._suds_object = suds_object
        else:
            self._get_suds_object()
        # initialize all instance attributes as properties
        # check if the property already exists. If so, use existing.
        for key in list(self._cls_suds_map.keys()):
            if not hasattr(self.__class__, key):
                # require default values for lambda or it evaluates all
                # variables
                # at the end of function (self._cls_suds_map[key] was evaluated
                # as the last key value in the for loop for all defined lambdas
                # Property Builder, parses _cls_suds_map to build properties:
                # custom fields:
                #    getter has parameters:
                #        field_name
                #    setter has parameters:
                #        val: the value that the property is set to
                #        field_name
                # array object fields:
                #    getter has parameters:
                #        field_name
                #    setter has parameters:
                #        val: the value that the property is set to
                #        field_name
                # object fields:
                #    getter has parameters:
                #        field_name
                #    setter has parameters:
                #        val: the value that the property is set to
                #        field_name
                # regular fields;
                #    use getattr and setattr
                if isinstance(self._cls_suds_map[key], dict):
                    if "is_custom" in self._cls_suds_map[key]:
                        setattr(self.__class__, key, property(
                            lambda self, field_name=key:
                                self._custom_getter(field_name),
                            lambda self, val, field_name=key:
                                self._custom_setter(val, field_name)))
                    elif "is_array" in self._cls_suds_map[key]:
                        setattr(self.__class__, key, property(
                            lambda self, field_name=key:
                                self._arr_obj_getter(field_name),
                            lambda self, val, field_name=key:
                                self._arr_obj_setter(val, field_name)))
                    else:
                        setattr(self.__class__, key, property(
                            lambda self, field_name=key:
                                self._obj_getter(field_name),
                            lambda self, val, field_name=key:
                                self._obj_setter(val, field_name)))
                else:
                    setattr(self.__class__, key, property(
                        # if the attribute doesn't exist in the current object
                        # return None
                        lambda self, suds_key=self._cls_suds_map[key]:
                            getattr(self._suds_object, suds_key, None),
                        lambda self, value, suds_key=self._cls_suds_map[key]:
                            self._regular_setter(value, suds_key)))
# after all properties are defined set the id field to the value passed in.
        if obj_id is not None:
            setattr(self, self._id_field, obj_id)

    def _get_suds_object(self):
        """Returns the WSDL object as created by the Polarion WSDL factory"""
        if self._obj_client and self._obj_struct:
            self._suds_object = getattr(self.session, self._obj_client). \
                factory.create(self._obj_struct)
        else:
            self._suds_object = None

    def _obj_getter(self, field_name):
        """get function for attributes that reference an object.
        Returns the referenced object. If the WSDL attribute contains a value
        that value is given to the object as its obj_id. Classes that have an
        _id_field attribute will return only that attribute when the property
        is gotten

        Args:
            field_name: the field name of the Polarion object to get
        """
        csm = self._cls_suds_map[field_name]
        named_arg = csm.get("named_arg", "suds_object")
        suds_field_val = getattr(
            self._suds_object, csm.get("field_name", ""), None)
        cls_obj = csm["cls"]
        if suds_field_val:
            if named_arg == "uri" and cls_obj._id_field:
                if suds_field_val:
                    id_re = re.search(self.REGEX_ID, suds_field_val)
                    if id_re:
                        return cls_obj.URI_ID_GET_REPLACE(id_re.group(1))
            else:
                args = {}
                args[named_arg] = suds_field_val
                obj = cls_obj(**args)
        else:
            obj = cls_obj()
        if cls_obj._id_field:
            return getattr(obj, obj._id_field)
        else:
            return obj

    def _obj_setter(self, val, field_name):
        """set function for attributes that reference an object. It can accept
        a string, a Pylero object or a raw WSDL object. If a string is given,
        it is passed in to the object as its obj_id.

        Args:
            val: the value that the property is being set to
            field_name: the field name of the Polarion object to set
        """
        csm = self._cls_suds_map[field_name]
        suds_field_name = csm["field_name"]
        enum_id = csm.get("enum_id")
        enum_override = csm.get("enum_override", [])
        sync_field = csm.get("sync_field")
        obj_cls = csm.get("cls")
        # deepcopy so that changes do not stick
        add_parms = copy.deepcopy(csm.get("additional_parms", {}))
        if isinstance(val, basestring):
            val = self._check_encode(val)
            if enum_id and val not in enum_override:
                self.check_valid_field_values(
                    val, enum_id, {},
                    self._wi_type if hasattr(self, "_wi_type") else None)
        if not sync_field:
            sync_field = "_suds_object"
        if isinstance(val, basestring) or val is None:
            add_parms[obj_cls._id_field] = val
            obj = obj_cls(**add_parms)
            setattr(self._suds_object, suds_field_name,
                    getattr(obj, sync_field))
        elif isinstance(val, obj_cls):
            setattr(self._suds_object, suds_field_name,
                    getattr(val, sync_field))
        elif isinstance(val, obj_cls()._suds_object.__class__):
            obj = obj_cls()
            if sync_field in obj._cls_suds_map:
                suds_sync_field = obj._cls_suds_map[sync_field]
                # if sync_field is given, the attribute will be simple
                val = getattr(val, suds_sync_field)
            setattr(self._suds_object, suds_field_name, val)
        else:
            raise PyleroLibException("the value {0} is not a valid type".
                                     format(val))

    def _arr_obj_getter(self, field_name):
        """get function for attributes that reference an array of objects.
        The Polarion array object always has a single Python list item which
        contains a list of the WSDL objects. This function converts each WSDL
        object to its Pylero object and returns that list

        Args:
            field_name: the field name of the Polarion object to get
        """
        csm = self._cls_suds_map[field_name]
        if getattr(self._suds_object, csm["field_name"], None):
            obj_lst = []
            # ArrayOf Polarion objects have a double list.
            for inst in getattr(self._suds_object, csm["field_name"])[0]:
                obj_lst.append(csm["cls"](suds_object=inst))
            return obj_lst
        else:
            return []

    def _arr_obj_setter(self, val, field_name):
        """set function for attributes that reference an array of objects. It
        requires a single instance or list of either Pylero or WSDL objects
        or an empty list.
        An empty list erases the attribute value.
        Otherwise it sets the attribute the value passed in.

        Args:
            val: the value that the property is set to
            field_name: the field name of the Polarion object to set
        """
        # TODO: Still needs to be fully tested. Looks like there are some bugs.
        csm = self._cls_suds_map[field_name]
        arr_inst = csm.get("arr_cls")()
        obj_inst = csm.get("cls")()
        # obj_attach =
        if not isinstance(val,
                          (list, arr_inst.__class__,
                           arr_inst._suds_object.__class__)):
            raise PyleroLibException(
                "{0}s must be a list of {1}").format(
                    csm["field_name"], obj_inst.__class__.__name__)
        elif not val:
            setattr(
                self._suds_object, csm["field_name"], arr_inst._suds_object)
        elif isinstance(val, arr_inst._suds_object.__class__):
            setattr(self._suds_object, csm["field_name"], val)
        elif isinstance(val, arr_inst.__class__):
            setattr(self._suds_object, csm["field_name"], val._suds_object)
        else:
            if isinstance(val, list):
                # if str values are based in, try instantiating a class with
                # the vals and then using that list. Then continue processing
                if isinstance(val[0], basestring):
                    val[0] = self._check_encode(val[0])
                    val = [csm["cls"](item) for item in val]

                if isinstance(val[0], obj_inst._suds_object.__class__):
                    setattr(getattr(self._suds_object, csm["field_name"]),
                            csm["inner_field_name"], val)
                else:
                    setattr(self._suds_object, csm["field_name"],
                            arr_inst._suds_object)
                    for item in val:
                        getattr(getattr(self._suds_object, csm["field_name"]),
                                csm["inner_field_name"]).append(
                                    item._suds_object)

    def custom_obj(self):
        # This returns a custom Polarion object. It can't use the Custom class
        # as that is a child of this class.
        return self.session.test_management_client.factory.create(
            "tns4:Custom")

    def custom_array_obj(self):
        # This returns a custom Polarion object. It can't use the Custom class
        # as that is a child of this class.
        return self.session.test_management_client.factory.create(
            "tns4:ArrayOfCustom")

    def _custom_getter(self, field_name):
        """Works with custom fields that has attributes stored differently
        then regular attributes. It first checks if there is a value in the
        local copy, which may have already been modified by the user. If not,
        it checks the server for a value.

        test_steps do not work like all other custom fields, and therefore
        require specific code. Its values are not saved in the local object.

        Args:
            field_name: the field name of the Polarion object to get
        """
        csm = self._cls_suds_map[field_name]
        if field_name == "test_steps":
            if self._changed_fields.get("testSteps"):
                return csm["cls"](
                    suds_object=self._changed_fields.get("testSteps"))
            else:
                test_steps = self.get_test_steps()
                if test_steps:
                    return test_steps
        else:
            if (("customFields" not in self._suds_object) or
                    (not self._suds_object.customFields)):
                self._suds_object.customFields = self.custom_array_obj()
            cf = self._suds_object.customFields[0]
            custom_fld = None
            if cf:
                # check if the custom field already exists and modify it.
                match = [x for x in cf if x.key == csm["field_name"]]
                if match:
                    custom_fld = match[0]
            if not custom_fld and self.uri:
                custom_fld = self.get_custom_field(
                    csm["field_name"])._suds_object
            if custom_fld:
                if isinstance(custom_fld, basestring):
                    obj = custom_fld
                elif csm.get("is_array"):
                    obj = []
                    # ArrayOf Polarion objects have a double list.
                    if custom_fld.value:
                        for inst in custom_fld.value[0]:
                            if csm["cls"]._cls_inner._id_field:
                                item_inst = csm["cls"]._cls_inner(
                                    suds_object=inst)
                                obj.append(
                                    getattr(item_inst, item_inst._id_field))
                            else:
                                obj.append(csm["cls"]._cls_inner(
                                    suds_object=inst))
                elif csm.get("cls"):
                    obj = csm["cls"](suds_object=custom_fld.value)
                else:
                    obj = custom_fld.value
                if getattr(obj, "_id_field", None):
                    return getattr(obj, obj._id_field)
                else:
                    return obj
            else:
                return None

    def _custom_setter(self, val, field_name):
        """Works with custom fields that has to keep track of values and what
        changed so that on update it can also update all the custom fields at
        the same time.

        Args:
            val: the value that the property is being set to
            field_name: the field name of the Polarion object to set
        """
        csm = self._cls_suds_map[field_name]
        if field_name == "test_steps":
            if not val:
                self._changed_fields[csm["field_name"]] = None
            elif isinstance(val, csm["cls"]):
                self._changed_fields[csm["field_name"]] = val._suds_object
            elif isinstance(val, csm["cls"]()._suds_object.__class__):
                self._changed_fields[csm["field_name"]] = val
            else:
                raise PyleroLibException(
                    "The value must be a {0}".format(csm["cls"].__name__))
        # move the custom fields to within the object, otherwise each custom
        # field is a seperate SVN commit. testSteps, does not work unless it
        # is uploaded using the set_test_steps function.
        else:
            cust = self.custom_obj()
            cust.key = csm["field_name"]
            if val is None:
                cust.value = None
            elif (not csm.get("cls") or isinstance(val, basestring)) \
                    and not csm.get("is_array"):
                # if there is no cls specified, val can be a bool, int, ...
                # if val is a string, it may be used to instantiate the class
                if isinstance(val, basestring):
                    val = self._check_encode(val)
                if csm.get("enum_id") and \
                        val not in csm.get("enum_override", []):
                    # uses deepcopy, to not affect other instances of the class
                    additional_parms = copy.deepcopy(
                        csm.get("additional_parms", {}))
                    self.check_valid_field_values(val, csm.get("enum_id"),
                                                  additional_parms,
                                                  csm.get("control"))
                cust.value = csm["cls"](val)._suds_object if csm.get("cls") \
                    else val
            elif csm.get("is_array"):
                if not isinstance(val, list):
                    raise PyleroLibException("value must be a list")
                if csm.get("enum_id"):
                    cust.value = csm["cls"]()._suds_object
                    for i in val:
                        if i not in csm.get("enum_override", []):
                            # uses deepcopy, to not affect other instances
                            # of the class
                            additional_parms = copy.deepcopy(
                                csm.get("additional_parms", {}))
                            self.check_valid_field_values(
                                i, csm.get("enum_id"), additional_parms,
                                self._wi_type if hasattr(self, "_wi_type")
                                else None)
                        cust.value[0].append(
                            csm["cls"]._cls_inner(i)._suds_object)

            elif isinstance(val, csm["cls"]):
                cust.value = val._suds_object
            elif isinstance(val, csm["cls"]()._suds_object.__class__):
                cust.value = val
            else:
                raise PyleroLibException(
                    "The value must be of type {0}."
                    .format(csm["cls"].__name__))
            if (("customFields" not in self._suds_object) or
                    (not self._suds_object.customFields)):
                self._suds_object.customFields = self.custom_array_obj()
            cf = self._suds_object.customFields[0]
            if cf:
                # check if the custom field already exists and modify it.
                match = [x for x in cf if x.key == csm["field_name"]]
                if match:
                    match[0].value = cust.value
                else:
                    cf.append(cust)
                    self._custom_fields = cf
            else:
                self._custom_fields = [cust]

    def _regular_setter(self, value, field_name):
        """This setter is used for any attributes that are not Polarion object
        data types. If the attribute type is a string, then it validates it
        using the check_encode function

        Args:
            value (string): the value that the property is being set to
            field_name: the field name of the Polarion object to set
        """
        if isinstance(value, basestring):
            value = self._check_encode(value)
        setattr(self._suds_object, field_name, value)

    def _check_encode(self, val):
        """Validate @val is a UTF-8 because Polarion doesn't support not
        UTF-8 characters. The only use case that is not taken into account
        is when an attribute is set directly with a SUDS object. The users
        have no way of calling this function in this case.

        Args:
        val (string): the value that the property is being set to
        """
        try:
            if not isinstance(val, type(u'')):
                val = val.decode('utf-8')
            # replace chr(160) with space
            return val.replace(u'\xa0', u' ')
        except UnicodeError as err:
            raise PyleroLibException(
                'String must be UTF-8 encoded. The following error was '
                'raised when converting it to unicode: {0}'
                .format(err)
            )

    def _get_file_data(self, path):
        """Method for getting attachment data that can be passed to the soap
        library. Is used by a number of child classes.

        Args:
            path: the file path

        Returns:
            base64 encoded binary data.
        """

        f = open(path, "rb")
        bdata = f.read()
        f.close()
        bencode = base64.b64encode(bdata)
        # Effectively turns in Python3 String
        try:
            # Python2
            return bencode.encode('utf8')
        except AttributeError:
            # Python3
            return bencode.decode('utf8')

    def _verify_obj(self):
        # verifies if the object contains a suds object from the server by
        # checking if the uri field is populated. If no URI it didn't come from
        # the server
        if not getattr(self, "uri", None):
            raise PyleroLibException("There is no {0} loaded".format(
                self.__class__.__name__))

    def can_add_element_to_key(self, key):
        """Checks if the current user can add elements to the collection at
        given key of the current object.

        Args:
            key: the key of the field that contains the collection.

        Returns:
            bool

        References:
            Security.canAddElementToKey
        """
        self._verify_obj()
        return self.session.security_client.service.canAddElementToKey(
            self.uri, key)

    def can_delete_instance(self):
        """Checks if the current user can delete the current object

        Args:
            None

        Returns:
            bool

        References:
            Security.canDeleteInstance
        """
        self._verify_obj()
        return self.session.security_client.service.canDeleteInstance(self.uri)

    def can_modify_instance(self):
        """Checks if the current user can modify the current object

        Args:
            None

        Returns:
            bool

        References:
            Security.canModifyInstance
        """
        self._verify_obj()
        return self.session.security_client.service.canModifyInstance(self.uri)

    def can_modify_key(self, key):
        """Checks if the current user can modify the field with given key of
        the current object.

        Args:
            key: the key of the field that contains the collection.

        Returns:
            bool

        References:
            Security.canModifyKey
        """
        self._verify_obj()
        return self.session.security_client.service.canModifyKey(self.uri, key)

    def can_read_instance(self):
        """Checks if the current user can read the current object

        Args:
            None

        Returns:
            bool

        References:
            Security.canReadInstance
        """
        self._verify_obj()
        return self.session.security_client.service.canReadInstance(self.uri)

    def can_read_key(self, key):
        """Checks if the current user can read the field with given key of
        the current object.

        Args:
            key: the key of the field that contains the collection.

        Returns:
            bool

        References:
            Security.canReadKey
        """
        self._verify_obj()
        return self.session.security_client.service.canReadKey(self.uri, key)

    def can_remove_element_from_key(self, key):
        """Checks if the current user can remove elements from the collection
        at given key of the current object.

        Args:
            key: the key of the field that contains the collection.

        Returns:
            bool

        References:
            Security.canRemoveElementFromKey
        """
        self._verify_obj()
        return self.session.security_client.service.canRemoveElementFromKey(
            self.uri, key)

    def get_location(self):
        """Returns the location of the current object. In the context of this
        service the method should be used to get the location of a
        project(-group).

        Args:
            None

        Returns:
            location (string)

        References:
            Security.getLocationForURI
        """
        self._verify_obj()
        return self.session.security_client.service.getLocationForURI(self.uri)

    def check_valid_field_values(self, val, enum_id, additional_parms,
                                 control=None):
        """verifies id the value passed in is valid for the enum or object
        passed in. for example, if we want to see if a valid user is given,
        this will try to instantiate the User class with the given parameter
        and additional parms. If it fails, it is not a valid value.

        Args:
            val: the value you want to set it to.
            enum_id: the enumeration or object to validate against
            additional_parms (dict): parms needed to instantiate class passed
                                    in as enum_id
            control: the control key for the enumeration. default:None
        """
        if isinstance(enum_id, type):
            try:
                # try to instantiate the object with the value and additional
                # parms. If that works, it is a valid value
                enum_id(val, **additional_parms)
            except Exception:
                raise PyleroLibException(
                    "{0} is not a valid value for {1}"
                    .format(val, enum_id.__name__))
        else:
            valid_values = self.get_valid_field_values(enum_id, control)
            if val not in valid_values:
                raise PyleroLibException("Acceptable values for {0} are:"
                                         "{1}".format(enum_id, valid_values))

    def get_valid_field_values(self, enum_id, control=None):
        """Gets the available enumeration options.
        Uses a cache dict because the time to get valid fields from server
        is time prohibitive.

        Args:
            enum_id: The enum code to get values for
            control: the control key for the enumeration. default:None

        Returns:
            Array of EnumOptions

        References:
            Tracker.getEnumOptionsForId
        """
        project_id = getattr(self, "project_id", None) or self.default_project
        enum_base = self._cache["enums"].get(enum_id)
        enums = None
        if enum_base:
            enums = enum_base.get(control)
        if not enums:
            enums = self.session.tracker_client.service. \
                getEnumOptionsForIdWithControl(project_id, enum_id, control)
            self._cache["enums"][enum_id] = {}
            self._cache["enums"][enum_id][control] = enums
        # the _cache contains _suds_object, so the id attribute is used.
        return [enum.id for enum in enums]

    def reload(self):
        """Reloads the object with data from the server.
        This function is useful if the data on the server changed or if a
        data changing function was called (such as TestRun.add_attachment)

        Notes:
            This will overwrite any unsaved data in the object.
        Args:
            None
        Returns:
            None
        """

        if getattr(self, "uri", None):
            obj = self.__class__(uri=self.uri)
            self._suds_object = obj._suds_object



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.properties import Properties


class TestsConfiguration(BasePolarion):
    """Object to handle the Polarion WSDL tns3:TestsConfiguration class

    Attributes:
        defect_auto_assignement_enabled (boolean)
        defect_reuse_type (string)
        defect_template (string)
        defect_to_test_case_link_role_id (string)
        defect_work_item_type (string)
        defects_project (string)
        fields_to_copy_from_test_case_to_defect (Properties)
        fields_to_copy_from_test_run_to_linked_defect (Properties)
        fields_to_copy_from_test_run_to_new_defect (Properties)
        max_created_defects (int)
        max_created_defects_percent (int)
        result_error_enum_id (string)
        result_failed_enum_id (string)
        result_passed_enum_id (string)
        retest_allowed (boolean)
        status_error_enum_id (string)
        status_failed_enum_id (string)
        status_ok_enum_id (string)
        summary_defect_severity (string)
        test_case_id_custom_field (string)
        test_case_template (string)
        test_case_test_comment_field_id (string)
        test_case_test_result_field_id (string)
        test_case_work_item_type (string)
        test_run_template (string)
"""
    _cls_suds_map = {"defect_auto_assignement_enabled":
                     "defectAutoAssignementEnabled",
                     "defect_reuse_type": "defectReuseType",
                     "defect_template": "defectTemplate",
                     "defect_to_test_case_link_role_id":
                     "defectToTestCaseLinkRoleId",
                     "defect_work_item_type": "defectWorkItemType",
                     "defects_project": "defectsProject",
                     "fields_to_copy_from_test_case_to_defect":
                     {"field_name": "fieldsToCopyFromTestCaseToDefect",
                      "cls": Properties},
                     "fields_to_copy_from_test_run_to_linked_defect":
                     {"field_name": "fieldsToCopyFromTestRunToLinkedDefect",
                      "cls": Properties},
                     "fields_to_copy_from_test_run_to_new_defect":
                     {"field_name": "fieldsToCopyFromTestRunToNewDefect",
                      "cls": Properties},
                     "max_created_defects": "maxCreatedDefects",
                     "max_created_defects_percent": "maxCreatedDefectsPercent",
                     "result_error_enum_id": "resultErrorEnumId",
                     "result_failed_enum_id": "resultFailedEnumId",
                     "result_passed_enum_id": "resultPassedEnumId",
                     "retest_allowed": "retestAllowed",
                     "status_error_enum_id": "statusErrorEnumId",
                     "status_failed_enum_id": "statusFailedEnumId",
                     "status_ok_enum_id": "statusOkEnumId",
                     "summary_defect_severity": "summaryDefectSeverity",
                     "test_case_id_custom_field": "testCaseIdCustomField",
                     "test_case_template": "testCaseTemplate",
                     "test_case_test_comment_field_id":
                     "testCaseTestCommentFieldId",
                     "test_case_test_result_field_id":
                     "testCaseTestResultFieldId",
                     "test_case_work_item_type": "testCaseWorkItemType",
                     "test_run_template": "testRunTemplate",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "test_management_client"
    _obj_struct = "tns3:TestsConfiguration"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class ActivityCustomValue(BasePolarion):
    """Object to handle the Polarion WSDL tns3:ActivityCustomValue class

    Attributes:
        values (ArrayOf_xsd_string)
"""
    _cls_suds_map = {"values": "values",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "tracker_client"
    _obj_struct = "tns3:ActivityCustomValue"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.user import User


class Attachment(BasePolarion):
    """Object to handle the Polarion WSDL tns5:Attachment class

    Attributes:
        author (User)
        file_name (string)
        attachment_id (string)
        length (long)
        title (string)
        updated (dateTime)
        url (string)
"""
    _cls_suds_map = {"author":
                     {"field_name": "author",
                      "cls": User},
                     "file_name": "fileName",
                     "attachment_id": "id",
                     "length": "length",
                     "title": "title",
                     "updated": "updated",
                     "url": "url",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns5:Attachment"


class ArrayOfAttachment(BasePolarion):
    _obj_client = "builder_client"
    _obj_struct = "tns5:ArrayOfAttachment"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.text import Text


class Category(BasePolarion):
    """Object to handle the Polarion WSDL tns5:Category class

    Attributes:
        description (Text)
        category_id (string)
        name (string)
"""
    _cls_suds_map = {"description":
                     {"field_name": "description",
                      "cls": Text},
                     "category_id": "id",
                     "name": "name",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns5:Category"
    _id_field = "category_id"


class ArrayOfCategory(BasePolarion):
    _obj_client = "builder_client"
    _obj_struct = "tns5:ArrayOfCategory"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.enum_option_id import EnumOptionId
from pylero.user import User


class Signature(BasePolarion):
    """Object to handle the Polarion WSDL tns4:Signature class

    Attributes:
        signed_by (User)
        signed_revision (string)
        signer_role (string)
        verdict (EnumOptionId)
        verdict_time (dateTime)
"""
    _cls_suds_map = {"verdict":
                     {"field_name": "verdict",
                      "cls": EnumOptionId},
                     "signed_revision": "signedRevision",
                     "verdict_time": "verdictTime",
                     "signer_role": "signerRole",
                     "signed_by":
                     {"field_name": "signedBy",
                      "cls": User}}
    _obj_client = "test_management_client"
    _obj_struct = "tns4:Signature"


class ArrayOfSignature(BasePolarion):
    _obj_client = "test_management_client"
    _obj_struct = "tns4:ArrayOfSignature"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class EnumOptionId(BasePolarion):
    """An object to manage Polarion TestManagement tns4:EnumOptionId"""
    _cls_suds_map = {"enum_id": "id"}
    _id_field = "enum_id"
    _obj_client = "test_management_client"
    _obj_struct = "tns4:EnumOptionId"

    def __init__(self, enum_id=None, suds_object=None):
        super(EnumOptionId, self).__init__(enum_id, suds_object)


class ArrayOfEnumOptionId(BasePolarion):
    """An object to manage Polarion TestManagement tns4:ArrayOfEnumOptionId"""
    _obj_client = "test_management_client"
    _obj_struct = "tns4:ArrayOfEnumOptionId"
    _cls_inner = EnumOptionId



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.text import Text


class TimePoint(BasePolarion):
    """Object to handle the Polarion WSDL tns5:TimePoint class

    Attributes:
        closed (boolean)
        description (Text)
        earliest_planned_start (date)
        time_point_id (string)
        name (string)
        time (date)
"""
    _cls_suds_map = {"closed": "closed",
                     "description":
                     {"field_name": "description",
                      "cls": Text},
                     "earliest_planned_start": "earliestPlannedStart",
                     "time_point_id": "id",
                     "name": "name",
                     "time": "time",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns5:TimePoint"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy

from pylero.base_polarion import BasePolarion
from pylero.category import Category
from pylero.custom_field_type import CustomFieldType
from pylero.exceptions import PyleroLibException
from pylero.tests_configuration import TestsConfiguration
from pylero.text import Text
from pylero.user import User


class Project(BasePolarion):
    """Object to handle the Polarion WSDL tns4:Project class

    Attributes:
        active (boolean)
        description (Text)
        finish (date)
        lead (User)
        location (str)
        lock_work_records_date (date)
        name (string)
        project_group (ProjectGroup)
        project_id (string)
        start (date)
        tracker_prefix (string)
"""
    _cls_suds_map = {"active": "active",
                     "description":
                     {"field_name": "description",
                      "cls": Text},
                     "finish": "finish",
                     "lead":
                     {"field_name": "lead",
                      "cls": User},
                     "location": "location",
                     "lock_work_records_date": "lockWorkRecordsDate",
                     "name": "name",
                     "project_group":
                     {"field_name": "projectGroupURI",
                      # cls is defined in _fix_circular_refs function
                      "named_arg": "uri",
                      "sync_field": "uri"},
                     "project_id": "id",
                     "start": "start",
                     "tracker_prefix": "trackerPrefix",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _id_field = "project_id"
    _obj_client = "project_client"
    _obj_struct = "tns2:Project"

    URI_STRUCT = "subterra:data-service:objects:/default/" \
                 "%(id)s${%(obj)s}%(id)s"

    @classmethod
    def get_context_roles(cls, location):
        """Returns the context (project) roles for the given location.

        Args:
            location: the location of the context (project/project group)

        Returns:
            list of roles

        References:
            Security.getContextRoles
        """
        return cls.session.security_client.service.getContextRoles(location)

    def __init__(self, project_id=None, suds_object=None, location=None,
                 uri=None):
        """Project constructor.

        Args:
            project_id: when given, the object is populated with the
                         Project data.
            suds_object: PolarionProject object. When given, the object
                          is populated by object data.
            location: the location of the Polarion project
            uri: the uri that references the PolarionProject

        Notes:
            Either project_id or suds_object or location or uri can be passed
            in or none of them. If none of the identifying parameters are
            passed in an empty object is created

        References:
            Project.getProject
            Project.getProjectatLocation
            Project.getProjectByURI
        """
        super(self.__class__, self).__init__(project_id, suds_object)
        if project_id:
            # if the project is already cached, make a deep copy and use it.
            # If not, get it and add it to the cache.
            project = self._cache["projects"].get(project_id)
            if project:
                self._suds_object = copy.deepcopy(project)
            else:
                self._suds_object = self.session.project_client.service. \
                    getProject(project_id)
                self._cache["projects"][project_id] = self._suds_object
        elif location:
            self._suds_object = self.session.project_client.service. \
                getProjectatLocation(location)
        elif uri:
            self._suds_object = self.session.project_client.service. \
                getProjectByURI(uri)
        if project_id or location or uri:
            if getattr(self._suds_object, "_unresolvable", True):
                raise PyleroLibException("The Project was not found.")

    def _fix_circular_refs(self):
        # The module references ProjectGroup, which references this class
        # This is not allowed, so the self reference is defined here.
        from pylero.project_group import ProjectGroup
        self._cls_suds_map["project_group"]["cls"] = ProjectGroup

    def get_categories(self):
        """ method get_categories retrieves a list of Category objects

        Args:
            None

        Returns:
            list of Category objects

        References:
            Tracker.getCategories
        """
        self._verify_obj()
        categories = []
        for suds_cat in self.session.tracker_client.service. \
                getCategories(self.project_id):
            categories.append(Category(suds_object=suds_cat))
        return categories

    def get_defined_custom_field_keys(self, work_item_type_id):
        """Gets all custom field keys defined for a workitem type in a project.

        Args:
            work_item_type_id: the workitem type ID.

        Returns:
            list of key for the project for the given workitem type

        References:
            Tracker.getDefinedCustomFieldkeys
        """
        self._verify_obj()
        return self.session.tracker_client.service.getDefinedCustomFieldkeys(
            self.project_id, work_item_type_id)

    def get_defined_custom_field_type(self, work_item_type_id, key):
        """method get_defined_custom_field_type gets custom field definition
        of a work item type for the given key.

        Args:
            work_item_type_id: the workitem type ID.
            key: The key of the custom field

        Returns:
            CustomFieldType object

        References:
            Tracker.getDefinedCustomFieldType
        """
        self._verify_obj()
        suds_custom = self.session.tracker_client.service. \
            getDefinedCustomFieldType(self._uri, work_item_type_id, key)
        return CustomFieldType(suds_object=suds_custom)

    def get_defined_custom_field_types(self, work_item_type_id):
        """method get_defined_custom_field_type gets custom field definition
        of a work item type all keys.

        Args:
            work_item_type_id: the workitem type ID.

        Returns:
            list of CustomFieldType object

        References:
            Tracker.getDefinedCustomFieldType
        """
        self._verify_obj()
        customs = []
        for suds_custom in self.session.tracker_client.service. \
                getDefinedCustomFieldType(self._uri, work_item_type_id):
            customs.append(CustomFieldType(suds_object=suds_custom))
        return customs

    def get_document_locations(self):
        """Gets the document locations (e.g. LiveDocuments) for a project.

        Args:
            None

        Returns:
            list of (string) document locations

        References:
            Tracker.getDocumentLocations
        """
        self._verify_obj()
        return self.session.tracker_client.service. \
            getDocumentLocations(self.project_id)

    def get_document_spaces(self):
        """Gets the Module/Document spaces for the project.

        Args:
            None

        Returns:
            list of (string) document spaces

        References:
            Tracker.getDocumentSpaces
        """
        self._verify_obj()
        return self.session.tracker_client.service. \
            getDocumentSpaces(self.project_id)

    def get_project_users(self):
        """Gets users of the project

        Args:
            None

        Returns:
            list of u.User objects

        References:
            Project.getProjectUsers
        """
        self._verify_obj()
        users = []
        for suds_user in self.session.project_client.service. \
                getProjectUsers(self.project_id):
            users.append(User(suds_object=suds_user))

    def get_test_steps_configuration(self):
        """method get_test_steps_configuration retrieves a list of the
        Test Steps configuration for the project

        Args:
            None

        Returns:
            list of configuration of the Test Steps custom field.

        References:
            TestManagement.getTestStepsConfiguration
        """
        self._verify_obj()
        config_steps = self.session.test_management_client.service. \
            getTestStepsConfiguration(self.project_id)
        return config_steps[0]

    def get_tests_configuration(self):
        """method get_tests_configuration retrieves the test management
        configuration for the project

        Args:
            None

        Returns:
            TestsConfiguration object

        References:
            TestManagement.getTestsConfiguration
        """
        self._verify_obj()
        tests_config = self.session.test_management_client.service. \
            getTestsConfiguration(self.project_id)
        return TestsConfiguration(suds_object=tests_config)

    def get_wiki_spaces(self):
        """Returns Wiki spaces from current project

        Args:
            None

        Returns:
            Array of string

        References:
            Tracker.getWikiSpaces
        """
        self._verify_obj()
        return self.session.tracker_client.service.getWikiSpaces(
            self.project_id)



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.enum_option_id import EnumOptionId


class WorkflowAction(BasePolarion):
    """Object to handle the Polarion WSDL tns3:WorkflowAction class

    Attributes:
        action_id (int)
        action_name (string)
        cleaned_features (ArrayOf_xsd_string)
        native_action_id (string)
        required_features (ArrayOf_xsd_string)
        suggested_features (ArrayOf_xsd_string)
        target_status (EnumOptionId)
        unavailability_message (string)
"""
    _cls_suds_map = {"action_id": "actionId",
                     "action_name": "actionName",
                     "cleaned_features": "cleaned_features",
                     "native_action_id": "nativeActionId",
                     "required_features": "requiredFeatures",
                     "suggested_features": "suggestedFeatures",
                     "target_status":
                     {"field_name": "targetStatus",
                      "cls": EnumOptionId},
                     "unavailability_message": "unavailabilityMessage",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "tracker_client"
    _obj_struct = "tns3:WorkflowAction"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.subterra_uri import ArrayOfSubterraURI
from pylero.subterra_uri import SubterraURI


class Revision(BasePolarion):
    """Object to handle the Polarion WSDL tns4:Revision class

    Attributes:
        author (string)
        created (dateTime)
        internal_commit (boolean)
        linked_work_item_uris (ArrayOfSubterraURI)
        message (string)
        name (string)
        repository_name (string)
"""
    _cls_suds_map = {"author": "author",
                     "created": "created",
                     "internal_commit": "internalCommit",
                     "linked_work_item_uris":
                     {"field_name": "linkedWorkItemURIs",
                      "is_array": True,
                      "cls": SubterraURI,
                      "arr_cls": ArrayOfSubterraURI,
                      "inner_field_name": "SubterraURI"},
                     "message": "message",
                     "name": "name",
                     "repository_name": "repositoryName",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns4:Revision"

    @classmethod
    def query(cls, query, sort="name", fields=["name"], query_uris=False):
        """Searches revisions

        Args:
            query: query, Lucene
            sort: Lucene sort string, default - name
            fields: list of field names to fill in the returned
                    Revision (can be null). For nested structures in
                    the lists you can use following syntax to include only
                    subset of fields: myList.LIST.key
                    (e.g. linkedWorkItems.LIST.role).
                    For custom fields you can specify which fields you want to
                    be filled using following syntax:
                    customFields.CUSTOM_FIELD_ID (e.g. customFields.risk).
                    Default - list containing "name"
            query_uris: if True, returns a list of URIs instead of Revision
                        objects. default - False

        Returns:
            list of Revisions

        References:
            Tracker.queryRevisions
        """
        if query_uris:
            return cls.session.tracker_client.service.queryRevisionUris(
                query, sort, False)
        else:
            revs = cls.session.tracker_client.service.queryRevisions(
                query, sort, fields)
            lst_rev = [Revision(suds_object=rev) for rev in revs]
            return lst_rev


class ArrayOfRevision(BasePolarion):
    _obj_client = "builder_client"
    _obj_struct = "tns4:ArrayOfRevision"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class Text(BasePolarion):
    """Object to manage Polarion TestManagement WS tns2:Text

    Rich text object

    Attributes:
        content (str): the formatted text
        content_type (str): indication of how content is formatted
                            (eg. text/html)
        content_lossy (bool):
    """
    _cls_suds_map = {"content": "content",
                     "content_type": "type",
                     "content_lossy": "contentLossy"}
    _id_field = "content"
    _obj_client = "test_management_client"
    _obj_struct = "tns2:Text"

    def __init__(self, content=None, suds_object=None):
        super(self.__class__, self).__init__(content, suds_object)

    def _get_suds_object(self):
        super(self.__class__, self)._get_suds_object()
        # default values
        self._suds_object.type = "text/html"
        self._suds_object.contentLossy = False


class ArrayOfText(BasePolarion):
    """Object containing list of Text objects"""
    _obj_client = "test_management_client"
    _obj_struct = "tns2:ArrayOfText"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.activity_custom_value import ActivityCustomValue
from pylero.base_polarion import BasePolarion


class ActivityCustomValueEntry(BasePolarion):
    """Object to handle the Polarion WSDL tns3:ActivityCustomValueEntry class

    Attributes:
        custom_values (ActivityCustomValue)
        key (string)
"""
    _cls_suds_map = {"custom_values":
                     {"field_name": "customValues",
                      "cls": ActivityCustomValue},
                     "key": "key",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "tracker_client"
    _obj_struct = "tns3:ActivityCustomValueEntry"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class EnumCustomFieldType(BasePolarion):
    """Object to handle the Polarion WSDL tns5:EnumCustomFieldType class

    Attributes:
        default_value (anyType)
        depends_on (string)
        description (string)
        enum_id (string)
        id (string)
        name (string)
        required (boolean)
        type (string)
"""
# id field is called cft_id and not enum_custom_field_type_id because it is
# often mixed with custom_field_type and they need to be accessed
# interchangeably
    _cls_suds_map = {"default_value": "defaultValue",
                     "depends_on": "dependsOn",
                     "description": "description",
                     "enum_id": "enumId",
                     "cft_id": "id",
                     "name": "name",
                     "multi": "multi",
                     "required": "required",
                     "type": "type",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns5:EnumCustomFieldType"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

"""Pylero is a Python wrapper for the Polarion WSDL API. It implements the
API in a native Python object oriented approach, using objects to wrap like
functionality.
"""



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.enum_option_id import EnumOptionId
from pylero.test_run_attachment import ArrayOfTestRunAttachment
from pylero.test_run_attachment import TestRunAttachment
from pylero.test_step_result import ArrayOfTestStepResult
from pylero.test_step_result import TestStepResult
from pylero.text import Text
from pylero.user import User
from pylero.work_item import _WorkItem


class TestRecord(BasePolarion):
    """Object to handle the Polarion WSDL tns3:TestRecord class

    Attributes (for specific details, see Polarion):
        attachments (ArrayOfTestRunAttachment)
        comment (Text)
        defect_case_id (string)
        duration (float)
        executed (dateTime)
        executed_by (string)
        result (EnumOptionId)
        test_case_revision (string)
        test_case_id (string)
        test_step_results (ArrayOfTestStepResult)
"""

    _cls_suds_map = {
        "attachments":
            {"field_name": "attachments",
             "is_array": True,
             "cls": TestRunAttachment,
             "arr_cls": ArrayOfTestRunAttachment,
             "inner_field_name": "TestRunAttachment"},
        "comment":
            {"field_name": "comment",
             "cls": Text},
        "defect_case_id":
            {"field_name": "defectURI",
             "cls": _WorkItem,
             "named_arg": "uri",
             "sync_field": "uri"},
        "duration": "duration",
        "executed": "executed",
        "executed_by":
            {"field_name": "executedByURI",
             "cls": User,
             "named_arg": "uri",
             "sync_field": "uri"},
        "result":
            {"field_name": "result",
             "cls": EnumOptionId,
             "enum_id": "result"},
        "test_case_id":
            {"field_name": "testCaseURI",
             "cls": _WorkItem,
             "named_arg": "uri",
             "sync_field": "uri"},
        "test_case_revision": "testCaseRevision",
        "test_step_results":
            {"field_name": "testStepResults",
             "is_array": True,
             "cls": TestStepResult,
             "arr_cls": ArrayOfTestStepResult,
             "inner_field_name": "TestStepResult"}}
    _obj_client = "test_management_client"
    _obj_struct = "tns3:TestRecord"
    _id_field = "test_case_id"

    def __init__(self, project_id=None, test_case_id=None, suds_object=None):
        self.project_id = project_id if project_id else self.default_project
        super(self.__class__, self).__init__(test_case_id, suds_object)

    def _fix_circular_refs(self):
        # need to pass in the project_id parm to the Work Item,
        # but it is not given before instatiation
        self._cls_suds_map["test_case_id"]["additional_parms"] = \
            {"project_id": self.project_id}
        self._cls_suds_map["defect_case_id"]["additional_parms"] = \
            {"project_id": self.project_id}


class ArrayOfTestRecord(BasePolarion):
    _obj_client = "test_management_client"
    _obj_struct = "tns3:ArrayOfTestRecord"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion
from pylero.enum_option_id import EnumOptionId
from pylero.user import User


class WorkRecord(BasePolarion):
    """Object to handle the Polarion WSDL tns5:WorkRecord class

    Attributes:
        comment (string)
        date (date)
        work_record_id (string)
        time_spent (duration)
        type (EnumOptionId)
        user (User)
"""
    _cls_suds_map = {"comment": "comment",
                     "date": "date",
                     "work_record_id": "id",
                     "time_spent": "timeSpent",
                     "type":
                     {"field_name": "type",
                      "cls": EnumOptionId,
                      "enum_id": "work-record-type"},
                     "user":
                     {"field_name": "user",
                      "cls": User},
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns5:WorkRecord"


class ArrayOfWorkRecord(BasePolarion):
    _obj_client = "builder_client"
    _obj_struct = "tns5:ArrayOfWorkRecord"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class Change(BasePolarion):
    """Object to handle the Polarion WSDL tns3:Change class

    Attributes:
        creation (boolean)
        date (dateTime)
        diffs (ArrayOf_tns3_FieldDiff)
        empty (boolean)
        invalid (boolean)
        revision (string)
        user (string)
    """
    _cls_suds_map = {"creation": "creation",
                     "date": "date",
                     "diffs": "diffs",
                     "empty": "empty",
                     "invalid": "invalid",
                     "revision": "revision",
                     "user": "user",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "tracker_client"
    _obj_struct = "tns3:Change"



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class FieldDiff(BasePolarion):
    """Object to handle the Polarion WSDL tns3:FieldDiff class

    Attributes:
        added (ArrayOf_xsd_anyType)
        after (anyType)
        before (anyType)
        collection (boolean)
        field_name (string)
        removed (ArrayOf_xsd_anyType)
"""
    _cls_suds_map = {"added": "added",
                     "after": "after",
                     "before": "before",
                     "collection": "collection",
                     "field_name": "fieldName",
                     "removed": "removed",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "tracker_client"
    _obj_struct = "tns3:FieldDiff"



import os
from distutils.command.install import INSTALL_SCHEMES

from setuptools import setup

with open("README.md", "r") as handle:
    LONG_DESCRIPTION = handle.read()

PACKAGE_NAME = "pylero"
CLI_NAME = "pylero-cmd"
RELEASE_FILE = "/etc/system-release-cpe"
SUDS_NAME_CHANGE = False

# change the data dir to be the etc dir under the package dir
for scheme in list(INSTALL_SCHEMES.values()):
    scheme['data'] = '%s/%s/etc' % (scheme['purelib'], PACKAGE_NAME)

install_requires_ = [
    'click',
]

# Update install_requires_ for feodra 36 and greater
if os.path.exists(RELEASE_FILE):
    with open(RELEASE_FILE) as version_file:
        version_file_content = version_file.read().split(":")
        if (version_file_content[3] == "fedora") and (
                int(version_file_content[4]) > 35):
            SUDS_NAME_CHANGE = True

if SUDS_NAME_CHANGE:
    install_requires_.append('suds')
else:
    install_requires_.append('suds-community')


if __name__ == "__main__":
    setup(
        name=PACKAGE_NAME,
        version='0.0.3',
        description="Python SDK for Polarion",
        long_description=LONG_DESCRIPTION,
        long_description_content_type='text/markdown',
        url="https://github.com/RedHatQE/pylero",
        author="%s Developers" % PACKAGE_NAME,
        author_email="szacks@redhat.com",
        license='MIT',
        package_dir={
            PACKAGE_NAME: 'src/%s' % PACKAGE_NAME,
        },
        packages=[
            PACKAGE_NAME,
            PACKAGE_NAME+'.cli',
        ],
        scripts=[
            'scripts/%s' % PACKAGE_NAME,
            'scripts/%s' % CLI_NAME,
        ],
        data_files=[
            ('', ['etc/%s/%s.cfg' % (PACKAGE_NAME, PACKAGE_NAME)])
        ],
        install_requires=install_requires_,
        classifiers=[
            'Development Status :: 5 - Production/Stable',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
            'Intended Audience :: Developers',      # Define that your audience are developers
            'Topic :: Software Development :: Build Tools',
            'License :: OSI Approved :: MIT License',   # Again, pick a license
            'Programming Language :: Python :: 3',      # Specify which pyhton versions that you want to support
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
          ]
    )



# -*- coding: utf8 -*-
from __future__ import print_function

import datetime
import os

from pylero._compatible import object
from pylero._compatible import str
from pylero.document import Document
from pylero.plan import Plan
from pylero.test_run import TestRun
from pylero.work_item import Requirement
from pylero.work_item import TestCase


class CmdList(object):
    ''' An object to manage the command of list'''

    def list_documents_by_query(self,
                                query):
        fields = ['document_id',
                  'document_name',
                  'author',
                  'created',
                  'updated',
                  'updated_by']
        doc_list = Document.query(query, False, fields)

        return doc_list

    def print_documents(self, docs):
        print('Created%7sAuthor%7sDocument' % ('', ''))
        print('-------%7s------%7s--------' % ('', ''))

        for doc in docs:
            print('%-14s %-11s %s' % (doc.created.strftime('%Y-%m-%d'),
                                      doc.author,
                                      doc.document_id))

    def list_workitems_in_doc(self, doc_name_with_space):
        if doc_name_with_space.find('/') < 0:
            print("Document format should be: 'space/document'")
            exit(1)

        doc = Document(Document.default_project, doc_name_with_space)
        fields = ['work_item_id',
                  'author',
                  'title',
                  'type',
                  'status',
                  'assignee',
                  'categories',
                  'comments',
                  'created',
                  'approvals',
                  'updated']
        self.workitem_list = doc.get_work_items(None, True, fields)
        return self.workitem_list

    def print_workitems(self, workitems):
        number = len(workitems)
        print('\nTotal workitems: %d' % number)
        print('Created%7sType%10sID%12sTitle' % ('', '', ''))
        print('---------%5s-------%7s-----%9s--------' % ('', '', ''))

        for wi in workitems:
            created = str(wi.created).split(' ')[0]
            print('%-13s %-13s %-13s %s' % (created, wi.type,
                                            wi.work_item_id, wi.title))

    def list_workitems_by_query(self, query, wi_type):
        fields = ['work_item_id',
                  'title',
                  'author',
                  'created']

        if wi_type in ["testcase", "TestCase"]:
            workitem_list = TestCase.query(query, fields)
        elif wi_type in ["requirement", "Requirement"]:
            workitem_list = Requirement.query(query, fields)
        elif wi_type == '':
            workitem_list = TestCase.query(query, fields) + \
                Requirement.query(query, fields)
        else:
            print("'%s' is invalid. Use testcase or requirement" % wi_type)
            exit(0)

        return workitem_list

    def print_steps_for_testcase(self, case_id):
        tc = TestCase(TestCase.default_project, case_id)
        steps = tc.get_test_steps()

        for step in steps.steps:
            stp = step.values[0].content
            exp = step.values[1].content
            print('TestStep       = %s' % stp)
            print('ExpectedResult = %s\n' % exp)

        if steps.steps is None:
            print('No step for this tesecase!')

    def print_links_for_requirement(self, req_id):
        req = Requirement(Requirement.default_project, req_id)
        print('ID%-12sRole' % (''))
        print('-------%7s------' % (''))

        for linked in req.linked_work_items_derived:
            print('%-13s %-13s' % (linked.work_item_id,
                                   linked.role))

    def print_links_for_testcase(self, case_id):
        tc = TestCase(TestCase.default_project, case_id)
        print('ID%-12sRole' % (''))
        print('-------%7s------' % (''))

        for linked in tc.linked_work_items:
            print('%-13s %-13s' % (linked.work_item_id,
                                   linked.role))

    def print_runs_by_query(self, query, is_template=False):
        query_ful = 'project.id:%s AND %s' % (TestRun.default_project, query)
        fields = ['query',
                  'created',
                  'test_run_id',
                  'select_test_cases_by',
                  'status',
                  'plannedin',
                  'assignee',
                  'author']

        st = TestRun.search(query_ful,
                            fields,
                            'created',
                            -1,
                            is_template)
        Object = ''
        if is_template:
            Object = 'Template'

        prestr = 'Created Time %8sAuthor %3sAssignee' % ('', '')
        latstr = '%sStatus %3sPlanID%10s%s' % ('', '', '', Object)
        preln = '------------%9s------%4s------' % ('', '')
        latln = '%2s--------%2s-------%9s--------' % ('', '', '')

        print('%s %s' % (prestr, latstr))
        print('%s %s' % (preln, latln))

        for tp in st:
            created_time = str(tp.created).split('.')[0]
            print('%-20s %-9s %-8s %-10s%-15s %s' % (created_time,
                                                     tp.author,
                                                     tp.assignee,
                                                     tp.status,
                                                     tp.plannedin,
                                                     tp.test_run_id))

    def print_templates_by_query(self, query):
        self.print_runs_by_query(query, True)

    def print_testcases_from_run(self, run):
        tr = TestRun(run, None, TestRun.default_project)
        print('(Only CaseID can be displayed when --run=$template)')
        print('List cases for: %s\n' % run)
        ttstr = ('Created Time %8sStatus %1sExecutedBy %2sCaseID' % ('',
                                                                     '',
                                                                     ''))
        lnstr = ('------------%9s------%2s----------%3s------' % ('', '', ''))
        print(ttstr)
        print(lnstr)

        for rec in tr.records:
            time = str(rec.executed).split('.')[0]
            print('%-21s%-9s%-12s%-10s' % (time,
                                           rec.result,
                                           rec.executed_by,
                                           rec.test_case_id))

    def print_plan_ids(self, query):
        pls = Plan.search(query,
                          sort='due_date',
                          limit=-1,
                          fields=['due_date',
                                  'name',
                                  'plan_id'])

        ttstr = ('Due Date%-5sPlan ID%-24sPlan Name' % ('', ''))
        lnstr = ('-----------  ---------- %-20s---------' % '')
        print(ttstr)
        print(lnstr)

        for pl in pls:
            print('%-12s %-30s %s' % (pl.due_date, pl.plan_id, pl.name))


class CmdUpdate(object):
    ''' An object to manage the command of update'''

    def update_all_case_results_for_run(self, run, result, user, comment):

        run = run.strip()
        tr = TestRun(run, None, TestRun.default_project)
        print('\nUpdate %s:' % run)

        if not comment:
            comment = ''

        print('Total records: %d' % len(tr.records))
        print('Updated Date Time    Result  CaseID')
        print('-------------------  ------  -----------')

        if user == 'None':
            user = TestRun.logged_in_user_id

        for rec in tr.records:
            rec.executed = datetime.datetime.now()
            rec.executed_by = user
            executed_str = str(rec.executed).split('.')[0]
            rec.result = result
            rec.comment = comment

            print('%-20s %-7s %s' % (executed_str,
                                     result,
                                     rec.test_case_id))
            tr.update_test_record_by_object(rec.test_case_id,
                                            rec)
        print('Done!')

    def update_all_case_results_for_runs(self, runs, result, user, comment):
        if runs.find(','):
            for run in runs.split(','):
                self.update_all_case_results_for_run(run, result,
                                                     user, comment)
        else:
            print("Please use comma ',' to seperate your runs!")

    def update_1_case_result_for_run(self,
                                     run,
                                     testcase,
                                     result,
                                     user,
                                     comment):

        if not comment:
            comment = ''

        tr = TestRun(run.strip(), None, TestRun.default_project)
        print('Update %s:' % run)

        if user == 'None':
            user = TestRun.logged_in_user_id

        is_found = False
        for rec in tr.records:
            if rec.test_case_id == testcase:
                is_found = True
                rec.executed = datetime.datetime.now()
                rec.executed_by = user
                rec.result = result
                rec.comment = comment

                tr.update_test_record_by_object(testcase, rec)
                print("%4sSet %s to %s (verdict comment: '%s')" % ('',
                                                                   testcase,
                                                                   result,
                                                                   comment))
                return 0

        if not is_found:
            print('Test case %s is not found in run.' % testcase)

    def update_document(self,
                        space,
                        doc_name,
                        doc_title,
                        wi_types=None,
                        doc_type=None,
                        structure_link_role="parent",
                        content=''):

        cl = CmdList()
        if cl.list_documents_by_query(doc_name):
            print("Exit - Found same name '%s/%s'" % (space, doc_name))
            return

        file_path = ""
        # the content could be file or data
        if os.path.exists(content):
            file_path = content
            with open(content, mode='r') as check_file:
                content = check_file.read()

        doc = Document.create(Document.default_project,
                              space,
                              doc_name,
                              doc_title,
                              wi_types,
                              doc_type,
                              structure_link_role,
                              content)

        if cl.list_documents_by_query(doc_name):
            print("Created document '%s/%s'" % (space, doc_name))
            print(" - document author      : %s" % doc.author)
            print(" - document type        : %s" % doc.type)
            print(" - allowed workitem type: %s" % wi_types)
            print(" - created date         : %s" % doc.created)
            if file_path != "":
                print(" - document content is from: %s" % file_path)
            else:
                print(" - document content     : %s" % content)
            return True
        else:
            print("Failed to create document '%s/%s'" % (space, doc_name))
            return False

    def update_run(self,
                   run,
                   template=None,
                   plannedin=None,
                   assignee=None,
                   status=None,
                   description=None,
                   is_template=False):

        run = run.strip()
        query_ful = 'project.id:%s AND id:\"%s\"' % (TestRun.default_project,
                                                     run)

        fields = ['query',
                  'created',
                  'test_run_id',
                  'select_test_cases_by',
                  'status',
                  'plannedin',
                  'assignee',
                  'author']
        st = TestRun.search(query_ful,
                            fields,
                            'created',
                            -1,
                            is_template)

        # Update run if exists, otherwise create it.
        if st:
            print('Update the existing run: %s' % run)
            tr = TestRun(run,
                         None,
                         TestRun.default_project)

            # set fields
            if assignee != 'None':
                tr.assignee = assignee
                print('%4sSet Assignee to %s' % ('', assignee))
            if plannedin is not None:
                tr.plannedin = plannedin
                print('%4sSet Plannedin to %s' % ('', plannedin))
            if status is not None:
                tr.status = status
                print('%4sSet Status to %s' % ('', status))
            if description is not None:
                tr.description = description
                print('%4sSet Description to %s' % ('', description))
            tr.update()

        else:
            tr = TestRun.create(TestRun.default_project,
                                run,
                                template,
                                assignee=assignee,
                                plannedin=plannedin,
                                status=status,
                                description=description)
            # display fields
            if assignee != 'None':
                print('%4sSet Assignee to %s' % ('', assignee))
            if plannedin is not None:
                print('%4sSet Plannedin to %s' % ('', plannedin))
            if status is not None:
                print('%4sSet Status to %s' % ('', status))
            if description is not None:
                print('%4sSet Description to %s' % ('', description))
            print('Created %s:' % run)

    def update_runs(self,
                    runs,
                    template=None,
                    plannedin=None,
                    assignee=None,
                    status=None,
                    description=None):

        if runs.find(','):
            for run in runs.split(','):
                self.update_run(run,
                                template,
                                plannedin,
                                assignee,
                                status,
                                description)
            print('Done!')
        else:
            print("Please use comma ',' to seperate your runs!")



# -*- coding: utf8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pylero.base_polarion import BasePolarion


class BuildTestResults(BasePolarion):
    """Object to handle the Polarion WSDL tns2:BuildTestResults class

    Attributes:
        error_count (int)
        failure_count (int)
        skipped_count (int)
        test_count (int)
"""
    _cls_suds_map = {"error_count": "errorCount",
                     "failure_count": "failureCount",
                     "skipped_count": "skippedCount",
                     "test_count": "testCount",
                     "uri": "_uri",
                     "_unresolved": "_unresolved"}
    _obj_client = "builder_client"
    _obj_struct = "tns2:BuildTestResults"
